import re
from os import PathLike
from pathlib import Path
from typing import Callable, List, Literal, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np

import emout.plot.basic_plot as emplt
import emout.utils as utils
from emout.plot.animation_plot import FrameUpdater, ANIMATER_PLOT_MODE
from emout.utils import DataFileInfo
import warnings


class Data(np.ndarray):
    """3次元データを管理する.

    Attributes
    ----------
    datafile : DataFileInfo
        データファイル情報
    name : str
        データ名
    slices : list(slice)
        管理するデータのxyz方向それぞれの範囲
    slice_axes : list(int)
        データ軸がxyzのどの方向に対応しているか表すリスト(0: t, 1: z, 2: y, 3: x)
    axisunits : list(UnitTranslator) or None
        軸の単位変換器
    valunit : UnitTranslator or None
        値の単位変換器
    """

    def __new__(
        cls,
        input_array,
        filename=None,
        name=None,
        xslice=None,
        yslice=None,
        zslice=None,
        tslice=None,
        slice_axes=None,
        axisunits=None,
        valunit=None,
    ):
        obj = np.asarray(input_array).view(cls)
        obj.datafile = DataFileInfo(filename)
        obj.name = name

        obj.axisunits = axisunits
        obj.valunit = valunit

        if xslice is None:
            xslice = slice(0, obj.shape[3], 1)
        if yslice is None:
            yslice = slice(0, obj.shape[2], 1)
        if zslice is None:
            zslice = slice(0, obj.shape[1], 1)
        if tslice is None:
            tslice = slice(0, obj.shape[0], 1)
        if slice_axes is None:
            slice_axes = [0, 1, 2, 3]

        obj.slices = [tslice, zslice, yslice, xslice]
        obj.slice_axes = slice_axes

        return obj

    def __getitem__(self, item):
        if not isinstance(item, tuple):
            item = (item,)

        new_obj = super().__getitem__(item)

        if not isinstance(new_obj, Data):
            return new_obj

        self.__add_slices(new_obj, item)

        params = {
            "filename": new_obj.filename,
            "name": new_obj.name,
            "xslice": new_obj.xslice,
            "yslice": new_obj.yslice,
            "zslice": new_obj.zslice,
            "tslice": new_obj.tslice,
            "slice_axes": new_obj.slice_axes,
            "axisunits": new_obj.axisunits,
            "valunit": new_obj.valunit,
        }

        if len(new_obj.shape) == 1:
            if isinstance(new_obj, Data1d):
                return new_obj
            return Data1d(new_obj, **params)
        elif len(new_obj.shape) == 2:
            if isinstance(new_obj, Data2d):
                return new_obj
            return Data2d(new_obj, **params)
        elif len(new_obj.shape) == 3:
            if isinstance(new_obj, Data3d):
                return new_obj
            return Data3d(new_obj, **params)
        elif len(new_obj.shape) == 4:
            if isinstance(new_obj, Data4d):
                return new_obj
            return Data4d(new_obj, **params)
        else:
            return new_obj

    def __add_slices(self, new_obj, item):
        """管理するデータの範囲を新しいオブジェクトに追加する.

        Parameters
        ----------
        new_obj : Data
            新しく生成されたデータオブジェクト
        item : int or slice or tuple(int or slice)
            スライス
        """
        slices = [*self.slices]
        axes = [*self.slice_axes]
        for i, axis in enumerate(axes):
            if i < len(item):
                slice_obj = item[i]
            else:
                continue

            if not isinstance(slice_obj, slice):
                slice_obj = slice(slice_obj, slice_obj + 1, 1)
                axes[i] = -1

            obj_start = slice_obj.start
            obj_stop = slice_obj.stop
            obj_step = slice_obj.step

            new_start = self.slices[axis].start
            new_stop = self.slices[axis].stop
            new_step = self.slices[axis].step

            if obj_start is not None:
                if obj_start < 0:
                    obj_start = self.shape[i] + obj_start
                new_start += self.slices[axis].step * obj_start

            if slice_obj.stop is not None:
                if obj_stop < 0:
                    obj_stop = self.shape[i] + obj_stop
                new_stop = self.slices[axis].start + self.slices[axis].step * obj_stop

            if obj_step is not None:
                new_step *= obj_step

            slices[axis] = slice(new_start, new_stop, new_step)

        axes = [axis for axis in axes if axis != -1]
        setattr(new_obj, "slices", slices)
        setattr(new_obj, "slice_axes", axes)

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.datafile = getattr(obj, "datafile", None)
        self.name = getattr(obj, "name", None)
        self.slices = getattr(obj, "slices", None)
        self.slice_axes = getattr(obj, "slice_axes", None)
        self.axisunits = getattr(obj, "axisunits", None)
        self.valunit = getattr(obj, "valunit", None)

    @property
    def filename(self) -> Path:
        """ファイル名を返す.

        Returns
        -------
        Path
            ファイル名.
        """
        return self.datafile.filename

    @property
    def directory(self) -> Path:
        """ディレクトリ名を返す

        Returns
        -------
        Path
            ディレクトリ名
        """
        return self.datafile.directory

    @property
    def xslice(self) -> slice:
        """管理するx方向の範囲を返す.

        Returns
        -------
        slice
            管理するx方向の範囲
        """
        return self.slices[3]

    @property
    def yslice(self) -> slice:
        """管理するy方向の範囲を返す.

        Returns
        -------
        slice
            管理するy方向の範囲
        """
        return self.slices[2]

    @property
    def zslice(self) -> slice:
        """管理するz方向の範囲を返す.

        Returns
        -------
        slice
            管理するz方向の範囲
        """
        return self.slices[1]

    @property
    def tslice(self) -> slice:
        """管理するt方向の範囲を返す.

        Returns
        -------
        slice
            管理するt方向の範囲
        """
        return self.slices[0]

    def axis(self, ax: int) -> np.ndarray:
        index = self.slice_axes[ax]
        axis_slice = self.slices[index]
        return np.array(*utils.slice2tuple(axis_slice))

    @property
    def x(self) -> np.ndarray:
        """x軸.

        Returns
        -------
        np.ndarray
            x軸
        """
        return np.arange(*utils.slice2tuple(self.xslice))

    @property
    def y(self) -> np.ndarray:
        """y軸.

        Returns
        -------
        np.ndarray
            y軸
        """
        return np.arange(*utils.slice2tuple(self.yslice))

    @property
    def z(self) -> np.ndarray:
        """z軸.

        Returns
        -------
        np.ndarray
            z軸
        """
        return np.arange(*utils.slice2tuple(self.zslice))

    @property
    def t(self) -> np.ndarray:
        """t軸.

        Returns
        -------
        np.ndarray
            t軸
        """
        slc = self.tslice
        maxlen = (slc.stop - slc.start) // slc.step
        return np.array(utils.range_with_slice(self.tslice, maxlen=maxlen))

    @property
    def x_si(self) -> np.ndarray:
        """SI単位系でのx軸.

        Returns
        -------
        np.ndarray
            SI単位系でのx軸
        """
        return self.axisunits[3].reverse(self.x)

    @property
    def y_si(self) -> np.ndarray:
        """SI単位系でのy軸.

        Returns
        -------
        np.ndarray
            SI単位系でのy軸
        """
        return self.axisunits[2].reverse(self.y)

    @property
    def z_si(self) -> np.ndarray:
        """SI単位系でのz軸.

        Returns
        -------
        np.ndarray
            SI単位系でのz軸
        """
        return self.axisunits[1].reverse(self.z)

    @property
    def t_si(self) -> np.ndarray:
        """SI単位系でのt軸.

        Returns
        -------
        np.ndarray
            SI単位系でのt軸
        """
        return self.axisunits[0].reverse(self.t)

    @property
    def val_si(self) -> "Data":
        """SI単位系での値.

        Returns
        -------
        Data
            SI単位系での値
        """
        return self.valunit.reverse(self)

    @property
    def use_axes(self) -> List[str]:
        """データ軸がxyzのどの方向に対応しているか表すリストを返す.

        Returns
        -------
        list(str)
            データ軸がxyzのどの方向に対応しているか表すリスト(['x'], ['x', 'z'], etc)
        """
        to_axis = {3: "x", 2: "y", 1: "z", 0: "t"}
        return list(map(lambda a: to_axis[a], self.slice_axes))

    def masked(
        self, mask: Union[np.ndarray, Callable[[np.ndarray], np.ndarray]]
    ) -> "Data":
        """マスクされたデータを返す.

        Parameters
        ----------
        mask : numpy.ndarray or predicate
            マスク行列またはマスクを返す関数

        Returns
        -------
        SlicedData
            マスクされたデータ
        """
        masked = self.copy()
        if isinstance(mask, np.ndarray):
            masked[mask] = np.nan
        else:
            masked[mask(masked)] = np.nan
        return masked

    def to_numpy(self) -> np.ndarray:
        """numpyのndarrayに変換する."""
        return np.array(self)

    def plot(self, **kwargs):
        """データをプロットする."""
        raise NotImplementedError()

    def build_frame_updater(
        self,
        axis: int = 0,
        title: Union[str, None] = None,
        notitle: bool = False,
        offsets: Union[
            Tuple[Union[float, str], Union[float, str], Union[float, str]], None
        ] = None,
        use_si: bool = True,
        vmin: float = None,
        vmax: float = None,
        **kwargs,
    ) -> FrameUpdater:
        """アニメーション描画処理を構築する.

        Parameters
        ----------
        axis : int, optional
            アニメーションする軸, by default 0
        title : str, optional
            タイトル(Noneの場合データ名(phisp等)), by default None
        notitle : bool, optional
            タイトルを付けない場合True, by default False
        offsets : (float or str, float or str, float or str)
            プロットのx,y,z軸のオフセット('left': 最初を0, 'center': 中心を0, 'right': 最後尾を0, float: 値だけずらす), by default None
        use_si : bool
            SI単位系を用いる場合True(そうでない場合EMSES単位系を用いる), by default False
        vmin : float, optional
            最小値, by default None
        vmax : float, optional
            最大値, by default None
        """
        updater = FrameUpdater(
            self, axis, title, notitle, offsets, use_si, vmin, vmax, **kwargs
        )

        return updater

    def gifplot(
        self,
        fig: Union[plt.Figure, None] = None,
        axis: int = 0,
        mode: str = None,
        action: ANIMATER_PLOT_MODE = "to_html",
        filename: PathLike = None,
        show: bool = False,
        savefilename: PathLike = None,
        interval: int = 200,
        repeat: bool = True,
        title: Union[str, None] = None,
        notitle: bool = False,
        offsets: Union[
            Tuple[Union[float, str], Union[float, str], Union[float, str]], None
        ] = None,
        use_si: bool = True,
        vmin: float = None,
        vmax: float = None,
        to_html: bool = False,
        return_updater: bool = False,
        **kwargs,
    ):
        """gifアニメーションを作成する.

        Parameters
        ----------
        fig : Figure
            アニメーションを描画するFigure(Noneの場合新しく作成する), by default None
        axis : int, optional
            アニメーションする軸, by default 0

        action : {'return', 'show', 'to_html', 'save', 'frames'}, optional, by default 'to_html'
            Determines the behavior of the function:

            - 'return': The plot object is returned without rendering it.
            - 'show': The plot is displayed immediately.
            - 'to_html': The plot is converted to an Ipython.display.HTML object and returned.
            - 'save': The plot is saved to a file specified by 'filename' argument.
            - 'frames': FrameUpdater object is returned without rendering it.

        filename : str, optional
            保存するファイル名(actionが'save'以外の場合やNoneの場合保存されない), by default None

        interval : int, optional
            フレーム間のインターバル(ミリ秒), by default 400
        repeat : bool
            アニメーションをループするならTrue, by default True
        title : str, optional
            タイトル(Noneの場合データ名(phisp等)), by default None
        notitle : bool, optional
            タイトルを付けない場合True, by default False
        offsets : (float or str, float or str, float or str)
            プロットのx,y,z軸のオフセット('left': 最初を0, 'center': 中心を0, 'right': 最後尾を0, float: 値だけずらす), by default None
        vmin : float, optional
            最小値, by default None
        vmax : float, optional
            最大値, by default None
        use_si : bool
            SI単位系を用いる場合True(そうでない場合EMSES単位系を用いる), by default False

        show : bool, optional
            プロットを表示する場合True(ファイルに保存する場合は非表示), by default

            .. deprecated :: 1.2.1

               This parameter is deprecated and will be removed in version 2.0.0.
               Use the 'action'='show' instead for equivalent functionality.

        savefilename : str, optional
            保存するファイル名(Noneの場合保存しない), by default None

            .. deprecated :: 1.2.1

               This parameter is deprecated and will be removed in version 2.0.0.
               Use the plot('action'='save', filename='example.gif') instead for equivalent functionality.

        to_html : bool
            アニメーションをHTMLとして返す. (使用例: Jupyter Notebook等でアニメーションを描画する際等)

            .. deprecated :: 1.2.1

               This parameter is deprecated and will be removed in version 2.0.0.
               Use the 'action'='to_html' instead for equivalent functionality.

        return_updater : bool
            FrameUpdaterを返す場合True, by default False

            .. deprecated :: 1.2.1

               This parameter is deprecated and will be removed in version 2.0.0.
               Use the 'action'='frames' instead for equivalent functionality.
               
        Returns
        -------
        Depending on the selected action:

        - If 'return': Returns the tuple of the plot object (fig, animation).
        - If 'show': Does not return anything, displays the plot.
        - If 'to_html': Returns an Ipython.display.HTML object of the plot (for Jupyter).
        - If 'save': Does not return anything, saves the plot to a file.
        - If 'frames': Returns FrameUpdater object.

        Examples
        --------
        >>> fig, ani = gifplot(action="return")
        Returns the tuple of the plot object.

        >>> gifplot(action="show")
        Displays the plot.

        >>> html = gifplot(action="to_html")
        Returns the HTML representation of the plot.

        >>> gifplot(action="save", filename = "example.gif")
        Saves the plot to a file.

        >>> updater = gifplot(action="frames")
        Returns FrameUpdater object.
        """
        if return_updater:
            warnings.warn(
                "The 'return_updater' flag is deprecated. "
                "Please use gifplot(action='frames') instead.",
                DeprecationWarning,
            )
            action = "frames"

        if mode is None:
            updater = self.build_frame_updater(
                axis, title, notitle, offsets, use_si, vmin, vmax, **kwargs
            )
        else:
            updater = self.build_frame_updater(
                axis, title, notitle, offsets, use_si, vmin, vmax, mode=mode, **kwargs
            )

        if action == "frames":
            return updater

        animator = updater.to_animator()

        return animator.plot(
            fig=fig,
            action=action,
            filename=filename,
            show=show,
            savefilename=savefilename,
            interval=interval,
            repeat=repeat,
            to_html=to_html,
            **kwargs,
        )


class Data4d(Data):
    """4次元データを管理する."""

    def __new__(cls, input_array, **kwargs):
        obj = np.asarray(input_array).view(cls)

        if "xslice" not in kwargs:
            kwargs["xslice"] = slice(0, obj.shape[3], 1)
        if "yslice" not in kwargs:
            kwargs["yslice"] = slice(0, obj.shape[2], 1)
        if "zslice" not in kwargs:
            kwargs["zslice"] = slice(0, obj.shape[1], 1)
        if "tslice" not in kwargs:
            kwargs["tslice"] = slice(0, obj.shape[0], 1)
        if "slice_axes" not in kwargs:
            kwargs["slice_axes"] = [0, 1, 2, 3]

        return super().__new__(cls, input_array, **kwargs)

    def plot(self, mode: Literal["auto"] = "auto", **kwargs):
        """4次元データをプロットする.(未実装)

        Parameters
        ----------
        mode : str, optional
            [description], by default 'auto'
        """
        if mode == "auto":
            mode = "".join(sorted(self.use_axes))
        pass


class Data3d(Data):
    """3次元データを管理する."""

    def __new__(cls, input_array, **kwargs):
        obj = np.asarray(input_array).view(cls)

        if "xslice" not in kwargs:
            kwargs["xslice"] = slice(0, obj.shape[2], 1)
        if "yslice" not in kwargs:
            kwargs["yslice"] = slice(0, obj.shape[1], 1)
        if "zslice" not in kwargs:
            kwargs["zslice"] = slice(0, obj.shape[0], 1)
        if "tslice" not in kwargs:
            kwargs["tslice"] = slice(0, 1, 1)
        if "slice_axes" not in kwargs:
            kwargs["slice_axes"] = [1, 2, 3]

        return super().__new__(cls, input_array, **kwargs)

    def plot(self, mode: Literal["auto"] = "auto", **kwargs):
        """3次元データをプロットする.(未実装)

        Parameters
        ----------
        mode : str, optional
            [description], by default 'auto'
        """
        if mode == "auto":
            mode = "".join(sorted(self.use_axes))
        pass


class Data2d(Data):
    """2次元データの2次元面を管理する."""

    def __new__(cls, input_array, **kwargs):
        obj = np.asarray(input_array).view(cls)

        if "xslice" not in kwargs:
            kwargs["xslice"] = slice(0, obj.shape[1], 1)
        if "yslice" not in kwargs:
            kwargs["yslice"] = slice(0, obj.shape[0], 1)
        if "zslice" not in kwargs:
            kwargs["zslice"] = slice(0, 1, 1)
        if "tslice" not in kwargs:
            kwargs["tslice"] = slice(0, 1, 1)
        if "slice_axes" not in kwargs:
            kwargs["slice_axes"] = [2, 3]

        return super().__new__(cls, input_array, **kwargs)

    def plot(
        self,
        axes: Literal["auto", "xy", "yz", "zx", "yx", "zy", "xy"] = "auto",
        show: bool = False,
        use_si: bool = True,
        offsets: Union[
            Tuple[Union[float, str], Union[float, str], Union[float, str]], None
        ] = None,
        mode: Literal["cm", "cm+cont", "cont"] = "cm",
        **kwargs,
    ):
        """2次元データをプロットする.

        Parameters
        ----------
        axes : str, optional
            プロットする軸('xy', 'zx', etc), by default 'auto'
        show : bool
            プロットを表示する場合True(ファイルに保存する場合は非表示), by default False
        use_si : bool
            SI単位系を用いる場合True(そうでない場合EMSES単位系を用いる), by default True
        offsets : (float or str, float or str, float or str)
            プロットのx,y,z軸のオフセット('left': 最初を0, 'center': 中心を0, 'right': 最後尾を0, float: 値だけずらす), by default None
        mode : str
            プロットの種類('cm': カラーマップ, 'cont': 等高線プロット, 'surf': サーフェースプロット)
        mesh : (numpy.ndarray, numpy.ndarray), optional
            メッシュ, by default None
        savefilename : str, optional
            保存するファイル名(Noneの場合保存しない), by default None
        cmap : matplotlib.Colormap or str or None, optional
            カラーマップ, by default cm.coolwarm
        vmin : float, optional
            最小値, by default None
        vmax : float, optional
            最大値, by default None
        figsize : (float, float), optional
            図のサイズ, by default None
        xlabel : str, optional
            x軸のラベル, by default None
        ylabel : str, optional
            y軸のラベル, by default None
        title : str, optional
            タイトル, by default None
        interpolation : str, optional
            用いる補間方法, by default 'bilinear'
        dpi : int, optional
            解像度(figsizeが指定された場合は無視される), by default 10

        Returns
        -------
        AxesImage or None
            プロットしたimageデータ(保存またはshowした場合None)

        Raises
        ------
        Exception
            プロットする軸のパラメータが間違っている場合の例外
        Exception
            プロットする軸がデータにない場合の例外
        Exception
            データの次元が2でない場合の例外
        """
        if self.valunit is None:
            use_si = False

        if axes == "auto":
            axes = "".join(sorted(self.use_axes))

        if not re.match(r"x[yzt]|y[xzt]|z[xyt]|t[xyz]", axes):
            raise Exception(
                'Error: axes "{axes}" cannot be used with Data2d'.format(axes=axes)
            )
        if axes[0] not in self.use_axes or axes[1] not in self.use_axes:
            raise Exception(
                'Error: axes "{axes}" cannot be used because {axes}-axis does not exist in this data.'.format(
                    axes=axes
                )
            )
        if len(self.shape) != 2:
            raise Exception(
                'Error: axes "{axes}" cannot be used because data is not 2dim shape.'.format(
                    axes=axes
                )
            )

        # x: 3, y: 2, z:1 t:0
        axis1 = self.slice_axes[self.use_axes.index(axes[0])]
        axis2 = self.slice_axes[self.use_axes.index(axes[1])]

        x = np.arange(*utils.slice2tuple(self.slices[axis1]))
        y = np.arange(*utils.slice2tuple(self.slices[axis2]))
        z = self if axis1 > axis2 else self.T  # 'xz'等の場合は転置

        if use_si:
            xunit = self.axisunits[axis1]
            yunit = self.axisunits[axis2]

            x = xunit.reverse(x)
            y = yunit.reverse(y)
            z = self.valunit.reverse(z)

            _xlabel = "{} [{}]".format(axes[0], xunit.unit)
            _ylabel = "{} [{}]".format(axes[1], yunit.unit)
            _title = "{} [{}]".format(self.name, self.valunit.unit)
        else:
            _xlabel = axes[0]
            _ylabel = axes[1]
            _title = self.name

        def _offseted(line, offset):
            if offset == "left":
                line -= line.ravel()[0]
            elif offset == "center":
                line -= line.ravel()[line.size // 2]
            elif offset == "right":
                line -= line.ravel()[-1]
            else:
                line += offset
            return line

        kwargs["xlabel"] = kwargs.get("xlabel", None) or _xlabel
        kwargs["ylabel"] = kwargs.get("ylabel", None) or _ylabel
        kwargs["title"] = kwargs.get("title", None) or _title

        if mode == "surf":
            mesh = np.meshgrid(x, y)

            kwargs["zlabel"] = kwargs.get("zlabel", None) or _title
            val = z
            if "x" not in self.use_axes:
                y, z = mesh
                x = self.x_si[0] if use_si else self.x[0]
                x = np.zeros_like(mesh[0]) + x
            elif "y" not in self.use_axes:
                x, z = mesh
                y = self.y_si[0] if use_si else self.y[0]
                y = np.zeros_like(mesh[0]) + y
            elif "z" not in self.use_axes:
                x, y = mesh
                z = self.z_si[0] if use_si else self.z[0]
                z = np.zeros_like(mesh[0]) + z

            if offsets is not None:
                x = _offseted(x, offsets[0])
                y = _offseted(y, offsets[1])
                z = _offseted(z, offsets[2])
                val = _offseted(val, offsets[3])

            imgs = [emplt.plot_surface(x, y, z, val, **kwargs)]
        else:
            if offsets is not None:
                x = _offseted(x, offsets[0])
                y = _offseted(y, offsets[1])
                z = _offseted(z, offsets[2])
            mesh = np.meshgrid(x, y)

            imgs = []
            if "cm" in mode and "cont" in mode:
                savefilename = kwargs.get("savefilename", None)
                kwargs["savefilename"] = None
                img = emplt.plot_2dmap(z, mesh=mesh, **kwargs)
                kwargs["savefilename"] = savefilename
                img2 = emplt.plot_2d_contour(z, mesh=mesh, **kwargs)
                imgs = [img, img2]
            elif "cm" in mode:
                img = emplt.plot_2dmap(z, mesh=mesh, **kwargs)
                imgs.append(img)
            elif "cont" in mode:
                img = emplt.plot_2d_contour(z, mesh=mesh, **kwargs)
                imgs.append(img)

        if show:
            plt.show()
            return None
        else:
            return imgs[0] if len(imgs) == 1 else imgs


class Data1d(Data):
    """3次元データの1次元直線を管理する."""

    def __new__(cls, input_array, **kwargs):
        obj = np.asarray(input_array).view(cls)

        if "xslice" not in kwargs:
            kwargs["xslice"] = slice(0, obj.shape[1], 1)
        if "yslice" not in kwargs:
            kwargs["yslice"] = slice(0, 1, 1)
        if "zslice" not in kwargs:
            kwargs["zslice"] = slice(0, 1, 1)
        if "tslice" not in kwargs:
            kwargs["tslice"] = slice(0, 1, 1)
        if "slice_axes" not in kwargs:
            kwargs["slice_axes"] = [3]

        return super().__new__(cls, input_array, **kwargs)

    def plot(
        self,
        show: bool = False,
        use_si: bool = True,
        offsets: Union[Tuple[Union[float, str], Union[float, str]], None] = None,
        **kwargs,
    ):
        """1次元データをプロットする.

        Parameters
        ----------
        show : bool
            プロットを表示する場合True(ファイルに保存する場合は非表示), by default False
        use_si : bool
            SI単位系を用いる場合True(そうでない場合EMSES単位系を用いる), by default True
        offsets : (float or str, float or str)
            プロットのx,y軸のオフセット('left': 最初を0, 'center': 中心を0, 'right': 最後尾を0, float: 値だけずらす), by default None
        savefilename : str, optional
            保存するファイル名, by default None
        vmin : float, optional
            最小値, by default None
        vmax : float, optional
            最大値, by default None
        figsize : (float, float), optional
            図のサイズ, by default None
        xlabel : str, optional
            横軸のラベル, by default None
        ylabel : str, optional
            縦軸のラベル, by default None
        label : str, optional
            ラベル, by default None
        title : str, optional
            タイトル, by default None

        Returns
        -------
        Line2D or None
            プロットデータを表す線オブジェクト(保存または show した場合None)

        Raises
        ------
        Exception
            データの次元が1でない場合の例外
        """
        if self.valunit is None:
            use_si = False

        if len(self.shape) != 1:
            raise Exception("Error: cannot plot because data is not 1dim shape.")

        axis = self.slice_axes[0]
        x = np.arange(*utils.slice2tuple(self.slices[axis]))
        y = self

        # "EMSES Unit" to "Physical Unit"
        if use_si:
            xunit = self.axisunits[axis]

            x = xunit.reverse(x)
            y = self.valunit.reverse(y)

            _xlabel = "{} [{}]".format(self.use_axes[0], xunit.unit)
            _ylabel = "{} [{}]".format(self.name, self.valunit.unit)
        else:
            _xlabel = self.use_axes[0]
            _ylabel = self.name

        def _offseted(line, offset):
            if offset == "left":
                line -= line[0]
            elif offset == "center":
                line -= line[len(line) // 2]
            elif offset == "right":
                line -= line[-1]
            else:
                line += offset
            return line

        if offsets is not None:
            x = _offseted(x, offsets[0])
            y = _offseted(y, offsets[1])

        kwargs["xlabel"] = kwargs.get("xlabel", None) or _xlabel
        kwargs["ylabel"] = kwargs.get("ylabel", None) or _ylabel

        line = emplt.plot_line(y, x=x, **kwargs)

        if show:
            plt.show()
            return None
        else:
            return line



import re

import f90nml
from numpy import isin

from emout.utils.units import Units


class UnitConversionKey:
    """単位変換キー.

    Attributes
    ----------
    dx : float
        グリッド幅[m]
    to_c : float
        EMSES単位系での光速の値
    """

    def __init__(self, dx, to_c):
        """単位変換キーを生成する.

        Parameters
        ----------
        dx : float
            グリッド幅[m]
        to_c : float
            EMSES単位系での光速の値
        """
        self.dx = dx
        self.to_c = to_c

    @classmethod
    def load(cls, filename):
        """ファイルから単位変換キーをロードする.

        ファイルの一行目に以下のような文字列が書かれている場合dx, to_cを読み取る.
        !!key dx=[1.0],to_c=[10000.0]

        Parameters
        ----------
        filename : str or Path
            単位変換キーを含むファイル.

        Returns
        -------
        UnitConversionKey or None
            単位変換キー
        """
        with open(filename, "r", encoding="utf-8") as f:
            line = f.readline()

        if not line.startswith("!!key"):
            return None

        # !!key dx=[1.0],to_c=[10000.0]
        text = line[6:].strip()
        pattern = r"dx=\[([+-]?\d+(?:\.\d+)?)\],to_c=\[([+-]?\d+(?:\.\d+)?)\]"
        m = re.match(pattern, text)
        dx = float(m.group(1))
        to_c = float(m.group(2))
        return UnitConversionKey(dx, to_c)

    @property
    def keytext(self):
        """単位変換キーの文字列を返す.

        Returns
        -------
        str
            単位変換キーの文字列
        """
        return "dx=[{}],to_c=[{}]".format(self.dx, self.to_c)


class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __getattr__(self, key):
        return self[key]


class InpFile:
    """パラメータファイルを管理する.

    nml : f90nml.Namelist
        Namelistオブジェクト
    """

    def __init__(self, filename=None, convkey=None):
        if filename:
            self.nml = f90nml.read(filename)
            self.convkey = UnitConversionKey.load(filename) or convkey
        else:
            self.nml = f90nml.Namelist()
            self.convkey = convkey

    def __contains__(self, key):
        if key in self.nml.keys():
            return True
        else:
            for group in self.nml.keys():
                if key in self.nml[group].keys():
                    return True
        return False

    def __getitem__(self, key):
        if key in self.nml.keys():
            return self.nml[key]
        else:
            for group in self.nml.keys():
                if key in self.nml[group].keys():
                    return self.nml[group][key]
        raise KeyError()

    def __setitem__(self, key, item):
        if key in self.nml.keys():
            self.nml[key] = item
            return
        else:
            for group in self.nml.keys():
                if key in self.nml[group].keys():
                    self.nml[group][key] = item
                    return
        raise KeyError()

    def __getattr__(self, key):
        item = self[key]
        if isinstance(item, dict):
            return AttrDict(item)
        return item

    def __setattr__(self, key, item):
        if key in ["nml", "convkey"]:
            super().__setattr__(key, item)
            return

        if key not in self:
            raise KeyError()

        self[key] = item

    def remove(self, key, index=None):
        """パラメータを削除する.

        Parameters
        ----------
        key : str
            グループ名(&groupname)またはパラメータ名(parameter)
        index : int, optional
            特定のインデックスのみ削除する場合指定する, by default None
        """
        if key in self.nml.keys():
            del self.nml[key]
        else:
            for group in self.nml.keys():
                if key in self.nml[group].keys():
                    try:
                        if index is None:
                            del self.nml[group][key]
                            del self.nml[group].start_index[key]
                        else:
                            (start_index,) = self.nml[group].start_index[key]
                            del self.nml[group][key][index - start_index]

                            if index == start_index:
                                self.nml[group].start_index[key] = [start_index + 1]

                            if len(list(self.nml[group][key])) == 0:
                                del self.nml[group][key]
                                del self.nml[group].start_index[key]
                    except (KeyError, IndexError):
                        pass
                    return

    def setlist(self, group, name, value, start_index=1):
        """リスト型のパラメータを設定する.

        Parameters
        ----------
        group : str
            グループ名(&groupname)
        name : str
            パラメータ名(parameter)
        value : type or list(type)
            設定する値
        start_index : int, optional
            設定するインデックス, by default 1
        """
        if not isinstance(value, list):
            value = [value]

        if name in self.nml[group].start_index:
            end_index = start_index + len(value)

            (start_index_init,) = self.nml[group].start_index[name]
            end_index_init = start_index_init + len(self.nml[group][name])

            min_start_index = min(start_index, start_index_init)
            max_end_index = max(end_index, end_index_init)

            new_list = [None] * (max_end_index - min_start_index)
            for i, index in enumerate(
                range(
                    start_index_init - min_start_index, end_index_init - min_start_index
                )
            ):
                new_list[index] = self.nml[group][name][i]
            for i, index in enumerate(
                range(start_index - min_start_index, end_index - min_start_index)
            ):
                new_list[index] = value[i]

            self.nml[group].start_index[name] = [min_start_index]
            self.nml[group][name] = new_list
        else:
            self.nml[group].start_index[name] = [start_index]
            self.nml[group][name] = value

    def save(self, filename, convkey=None):
        """パラメータをファイルに保存する.

        Parameters
        ----------
        filename : str or Path
            保存するファイル名
        convkey : UnitConversionKey, optional
            単位変換キー, by default None
        """
        convkey = convkey or self.convkey
        with open(filename, "wt", encoding="utf-8") as f:
            if convkey is not None:
                f.write("!!key {}\n".format(convkey.keytext))
            f90nml.write(self.nml, f, force=True)

    def __str__(self):
        return str(self.nml)

    def __repr__(self):
        return str(self)

    def conversion(self, unit_from: Units, unit_to: Units):
        def conv(group, name, unit_name):
            value_from = self[group][name]

            value = getattr(unit_from, unit_name).reverse(value_from)
            value_to = getattr(unit_to, unit_name).trans(value)

            self[group][name] = value_to

        def conv1d(group, name, unit_name):
            if group not in self:
                return
            if name not in self[group]:
                return

            values_from = self[group][name]

            values_to = []
            for value_from in values_from:
                if value_from is None:
                    value_to = None
                else:
                    value = getattr(unit_from, unit_name).reverse(value_from)
                    value_to = getattr(unit_to, unit_name).trans(value)
                values_to.append(value_to)

            self[group][name] = values_to

        conv1d("plasma", "wp", "f")

        conv1d("intp", "path", "v")
        conv1d("intp", "peth", "v")
        conv1d("intp", "vdri", "v")

        conv1d("emissn", "curf", "J")
        conv1d("emissn", "curfs", "J")

        self.convkey = UnitConversionKey(unit_to.dx, unit_to.to_c)

    @property
    def dx(self) -> float:
        return self.convkey.dx

    @property
    def to_c(self) -> float:
        return self.convkey.to_c


import math


class Group:
    """複数のオブジェクトをまとめて処理するクラス.

    Attributes
    ----------
    objs : list[Any]
        オブジェクトのリスト

    Notes
    -----
    以下の関数以外の組み込み関数は管理するオブジェクトに対して実行する.
    - repr : return 'Group([objs])
    - str : return 'Group([objs])'
    - format : return 'Group([objs])'
    - len : return len(objs)
    - iter : return iter(objs)
    - in : return other in objs

    Examples
    --------
    以下のように複数のオブジェクトをまとめて処理することができる.

    >>> group = Group([[1, 2], [3, 4]])
    >>> print(group)
    Group([[1, 2], [3, 4]])
    >>> print(group[0])
    Group([1, 3])
    >>> group.append(5)
    >>> print(group)
    Group([[1, 2, 5], [3, 4, 5]])
    """

    def __init__(self, objs, attrs=None):
        """グループを生成する.

        Parameters
        ----------
        objs : list
            オブジェクトのリスト
        """
        self.__dict__ = dict()
        self.objs = objs
        self.attrs = attrs

    def __binary_operator(self, callable, other):
        """グループに対して二項演算を行う.

        Parameters
        ----------
        callable : Callable[[Any, Any], Any]
            二項演算
        other : Any or Group
            対象(Groupの場合は管理するオブジェクトそれぞれに対して演算を適用する)

        Returns
        -------
        Group
            二項演算結果のグループ
        """
        if isinstance(other, Group):
            others = other
            new_objs = [callable(obj, other) for obj, other in zip(self.objs, others)]
        else:
            new_objs = [callable(obj, other) for obj in self.objs]
        return type(self)(new_objs, attrs=self.attrs)

    def __check_and_return_iterable(self, arg):
        """対象がGroupオブジェクトか判定し、対象をself.objs長にブロードキャストする.

        対象がGroupの場合:
            サイズが一致するかどうか検証し、そのままの対象を返す.
        対象がGroupでない場合:
            対象を複製し、self.objs長のリストにして返す.

        Parameters
        ----------
        arg : Any or Group
            チェックする対象

        Returns
        -------
        iterable (Group or list)
            self.objs長のイテラブルオブジェクト

        Raises
        ------
        Exception
            対象がGroupオブジェクトかつそのサイズが一致しない場合の例外
        """
        if isinstance(arg, Group):
            if len(self) != len(arg):
                raise Exception(
                    "Error: group size is not match (self:{}, arg:{})".format(
                        len(self), len(arg)
                    )
                )
            args = arg
        else:
            args = [arg] * len(self.objs)
        return args

    def map(self, callable):
        """すべてのオブジェクトに対して関数を適用し、新しいグループを生成する.

        Parameters
        ----------
        callable : Callable[Any, Any]
            適用する関数

        Returns
        -------
        Group
            新しいグループ

        Examples
        --------
        >>> group = Group([1, 2, 3])
        >>> group.map(lambda n: n * 2)
        Group([2, 4, 6])
        """
        new_objs = list(map(callable, self.objs))
        return type(self)(new_objs, attrs=self.attrs)

    def filter(self, predicate):
        """オブジェクトのうち関数が真を返すもののみで新しいグループを生成する.

        Parameters
        ----------
        predicate : Callable[Any, bool]
            フィルタ関数

        Returns
        -------
        Group
            新しいグループ

        Examples
        --------
        >>> group = Group([1, 2, 3])
        >>> group.filter(lambda n: n <= 2)
        Group([1, 2])
        """
        new_objs = list(filter(predicate, self.objs))
        return type(self)(new_objs, attrs=self.attrs)

    def foreach(self, callable):
        """すべてのオブジェクトに対して関数を適用する.

        Parameters
        ----------
        callable : Callable[Any, ]
            適用する関数

        Examples
        --------
        >>> group = Group([1, 2, 3])
        >>> group.foreach(print)
        1
        2
        3
        """
        for obj in self.objs:
            callable(obj)

    def __str__(self):
        return "Group({})".format(self.objs)

    def __repr__(self):
        return str(self)

    def __format__(self, format_spec):
        return str(self)

    def __len__(self):
        return len(self.objs)

    def __iter__(self):
        return iter(self.objs)

    def __contains__(self, obj):
        return obj in self.objs

    def __pos__(self):
        return self.map(lambda obj: -obj)

    def __neg__(self):
        return self.map(lambda obj: -obj)

    def __invert__(self):
        return self.map(lambda obj: ~obj)

    def __add__(self, other):
        return self.__binary_operator(lambda obj, other: obj + other, other)

    def __radd_(self, other):
        return self.__binary_operator(lambda obj, other: other + obj, other)

    def __sub__(self, other):
        return self.__binary_operator(lambda obj, other: obj - other, other)

    def __rsub(self, other):
        return self.__binary_operator(lambda obj, other: other - obj, other)

    def __mul__(self, other):
        return self.__binary_operator(lambda obj, other: obj * other, other)

    def __rmul__(self, other):
        return self.__binary_operator(lambda obj, other: other * obj, other)

    def __truediv__(self, other):
        return self.__binary_operator(lambda obj, other: obj / other, other)

    def __rtruediv__(self, other):
        return self.__binary_operator(lambda obj, other: other / obj, other)

    def __floordiv__(self, other):
        return self.__binary_operator(lambda obj, other: obj // other, other)

    def __rfloordiv__(self, other):
        return self.__binary_operator(lambda obj, other: other // obj, other)

    def __mod__(self, other):
        return self.__binary_operator(lambda obj, other: obj % other, other)

    def __rmod__(self, other):
        return self.__binary_operator(lambda obj, other: other % obj, other)

    def __divmod__(self, other):
        return self.__binary_operator(lambda obj, other: divmod(obj, other), other)

    def __rdivmod__(self, other):
        return self.__binary_operator(lambda obj, other: divmod(other, obj), other)

    def __pow__(self, other):
        return self.__binary_operator(lambda obj, other: obj**other, other)

    def __rpow__(self, other):
        return self.__binary_operator(lambda obj, other: other**obj, other)

    def __lshift__(self, other):
        return self.__binary_operator(lambda obj, other: obj << other, other)

    def __rlshift__(self, other):
        return self.__binary_operator(lambda obj, other: other << obj, other)

    def __rshift__(self, other):
        return self.__binary_operator(lambda obj, other: obj >> other, other)

    def __rrshift__(self, other):
        return self.__binary_operator(lambda obj, other: other >> obj, other)

    def __and__(self, other):
        return self.__binary_operator(lambda obj, other: obj & other, other)

    def __rand__(self, other):
        return self.__binary_operator(lambda obj, other: other & obj, other)

    def __or__(self, other):
        return self.__binary_operator(lambda obj, other: obj | other, other)

    def __ror__(self, other):
        return self.__binary_operator(lambda obj, other: other | obj, other)

    def __xor__(self, other):
        return self.__binary_operator(lambda obj, other: obj ^ other, other)

    def __rxor__(self, other):
        return self.__binary_operator(lambda obj, other: other ^ obj, other)

    def __abs__(self):
        return self.map(abs)

    def __eq__(self, other):
        return self.__binary_operator(lambda obj, other: obj == other, other)

    def __ne__(self, other):
        return self.__binary_operator(lambda obj, other: obj != other, other)

    def __le__(self, other):
        return self.__binary_operator(lambda obj, other: obj <= other, other)

    def __ge__(self, other):
        return self.__binary_operator(lambda obj, other: obj >= other, other)

    def __lt__(self, other):
        return self.__binary_operator(lambda obj, other: obj < other, other)

    def __gt__(self, other):
        return self.__binary_operator(lambda obj, other: obj > other, other)

    def __int__(self):
        return self.map(int)

    def __float__(self):
        return self.map(float)

    def __complex__(self):
        return self.map(complex)

    def __bool__(self):
        return self.map(bool)

    def __bytes__(self):
        return self.map(bytes)

    def __hash__(self):
        return self.map(hash)

    def __slice_to_iterable(self, slc):
        start = slc.start
        stop = slc.stop
        step = slc.step
        starts = self.__check_and_return_iterable(start)
        stops = self.__check_and_return_iterable(stop)
        steps = self.__check_and_return_iterable(step)
        slices = []
        for start, stop, step in zip(starts, stops, steps):
            slices.append(slice(start, stop, step))
        return slices

    def __expand_key(self, key):
        if not isinstance(key, tuple):
            key = (key,)

        slices = []
        for k in key:
            if isinstance(k, slice):
                slices.append(self.__slice_to_iterable(k))
            else:
                it = self.__check_and_return_iterable(k)
                slices.append(it)

        keys = []
        for key in zip(*slices):
            if len(key) == 1:
                keys.append(key[0])
            else:
                keys.append(tuple(key))

        return keys

    def __getitem__(self, key):
        keys = self.__expand_key(key)

        new_objs = [obj[key] for obj, key in zip(self.objs, keys)]

        return type(self)(new_objs, attrs=self.attrs)

    def __setitem__(self, key, value):
        keys = self.__expand_key(key)
        values = self.__check_and_return_iterable(value)

        for obj, key, value in zip(self.objs, keys, values):
            obj[key] = value

    def __delitem__(self, key):
        keys = self.__expand_key(key)

        for obj, key in zip(self.objs, keys):
            del obj[key]

    def __getattr__(self, key):
        keys = self.__check_and_return_iterable(key)
        new_objs = [getattr(obj, key) for obj, key in zip(self.objs, keys)]
        return type(self)(new_objs, attrs=self.attrs)

    def __setattr__(self, key, value):
        if key in ("objs", "__dict__", "attrs"):
            self.__dict__[key] = value
            return

        keys = self.__check_and_return_iterable(key)
        values = self.__check_and_return_iterable(value)

        for obj, key, value in zip(self.objs, keys, values):
            setattr(obj, key, value)

    def __call__(self, *args, **kwargs):
        new_argss = [list() for i in range(len(self.objs))]
        for arg in args:
            arg = self.__check_and_return_iterable(arg)
            for i, a in enumerate(arg):
                new_argss[i].append(a)

        new_kwargss = [dict() for i in range(len(self.objs))]
        for key, value in kwargs.items():
            keys = self.__check_and_return_iterable(key)
            values = self.__check_and_return_iterable(value)
            for i, (key, value) in enumerate(zip(keys, values)):
                new_kwargss[i][key] = value

        new_objs = [
            obj(*new_args, **new_kwargs)
            for obj, new_args, new_kwargs in zip(self.objs, new_argss, new_kwargss)
        ]

        return type(self)(new_objs, attrs=self.attrs)

    def __round__(self):
        return self.map(math.round)

    def __trunc__(self):
        return self.map(math.trunc)

    def __floor__(self):
        return self.map(math.floor)

    def __ceil__(self):
        return self.map(math.ceil)



from pathlib import Path
from typing import List

from .emout import Emout


def search_dirs(filename_patterns: List[str]) -> List[Path]:
    dirs: List[Path] = []
    for d in filename_patterns:
        dirs += Path().cwd().glob(d)
    return list(filter(lambda d: d.is_dir(), dirs))


def load_emouts(filename_patterns: List[str]) -> List[Emout]:
    dirs = search_dirs(filename_patterns)
    datas = []
    for d in dirs:
        data = Emout(d)
        datas.append(data)
    return datas



from itertools import chain
from os import PathLike
from pathlib import Path
from typing import List, Tuple, Union

import h5py
import numpy as np

import emout.utils as utils
from emout.utils import DataFileInfo, UnitTranslator

from .data import Data3d, Data4d


class GridDataSeries:
    """3次元時系列データを管理する.

    Attributes
    ----------
    datafile : DataFileInfo
        データファイル情報
    h5 : h5py.File
        hdf5ファイルオブジェクト
    group : h5py.Datasets
        データセット
    name : str
        データセット名
    """

    def __init__(
        self,
        filename: PathLike,
        name: str,
        tunit: UnitTranslator = None,
        axisunit: UnitTranslator = None,
        valunit: UnitTranslator = None,
    ):
        """3次元時系列データを生成する.

        Parameters
        ----------
        filename : str or Path
            ファイル名
        name : str
            データの名前
        """
        self.datafile = DataFileInfo(filename)
        self.h5 = h5py.File(str(filename), "r")
        self.group = self.h5[list(self.h5.keys())[0]]
        self._index2key = {int(key): key for key in self.group.keys()}
        self.tunit = tunit
        self.axisunit = axisunit
        self.valunit = valunit

        self.name = name

    def close(self) -> None:
        """hdf5ファイルを閉じる."""
        self.h5.close()

    def time_series(self, x, y, z) -> np.ndarray:
        """指定した範囲の時系列データを取得する.

        Parameters
        ----------
        x : int or slice
            x座標
        y : int or slice
            y座標
        z : int or slice
            z座標

        Returns
        -------
        numpy.ndarray
            指定した範囲の時系列データ
        """
        series = []
        indexes = sorted(self._index2key.keys())
        for index in indexes:
            key = self._index2key[index]
            series.append(self.group[key][z, y, x])
        return np.array(series)

    @property
    def filename(self) -> Path:
        """ファイル名を返す.

        Returns
        -------
        Path
            ファイル名
        """
        return self.datafile.filename

    @property
    def directory(self) -> Path:
        """ディレクトリ名を返す.

        Returns
        -------
        Path
            ディレクトリ名
        """
        return self.datafile.directory

    def _create_data_with_index(self, index: int) -> Data3d:
        """時間が指定された場合に、その時間におけるData3dを生成する.

        Parameters
        ----------
        index : int
            時間インデックス

        Returns
        -------
        Data3d
            生成したData3d

        Raises
        ------
        IndexError
            指定した時間が存在しない場合の例外
        """
        if index not in self._index2key:
            raise IndexError()

        key = self._index2key[index]

        axisunits = [self.tunit] + [self.axisunit] * 3

        return Data3d(
            np.array(self.group[key]),
            filename=self.filename,
            name=self.name,
            axisunits=axisunits,
            valunit=self.valunit,
        )

    def __create_data_with_indexes(
        self, indexes: List[int], tslice: slice = None
    ) -> Data4d:
        """時間が範囲で指定された場合に、Data4dを生成する.

        Parameters
        ----------
        indexes : list
            時間インデックスのリスト
        tslice : slice, optional
            時間インデックスの範囲, by default None

        Returns
        -------
        Data4d
            生成したData4d
        """
        if tslice is not None:
            start = tslice.start or 0
            stop = tslice.stop or len(self)
            step = tslice.step or 1
            tslice = slice(start, stop, step)

        array = []
        for i in indexes:
            array.append(self[i])

        axisunits = [self.tunit] + [self.axisunit] * 3

        return Data4d(
            np.array(array),
            filename=self.filename,
            name=self.name,
            tslice=tslice,
            axisunits=axisunits,
            valunit=self.valunit,
        )

    def __getitem__(
        self, item: Union[int, slice, List[int], Tuple[Union[int, slice, List[int]]]]
    ) -> Union["Data3d", "Data4d"]:
        """時系列データをスライスしたものを返す.

        Parameters
        ----------
        item : int or slice or list or tuple(int or slice or list)
            tzxyインデックスの範囲

        Returns
        -------
        Data3d or Data4d
            スライスされたデータ

        Raises
        ------
        TypeError
            itemのタイプが正しくない場合の例外
        """
        # xyzの範囲も指定された場合
        if isinstance(item, tuple):
            if isinstance(item[0], int):
                return self[item[0]][item[1:]]
            else:
                slices = (slice(None), *item[1:])
                return self[item[0]][slices]

        # 以下、tの範囲のみ指定された場合
        if isinstance(item, int):  # tが一つだけ指定された場合
            index = item
            if index < 0:
                index = len(self) + index
            return self._create_data_with_index(index)

        elif isinstance(item, slice):  # tがスライスで指定された場合
            indexes = list(utils.range_with_slice(item, maxlen=len(self)))
            return self.__create_data_with_indexes(indexes, tslice=item)

        elif isinstance(item, list):  # tがリストで指定された場合
            return self.__create_data_with_indexes(item)

        else:
            raise TypeError()

    def chain(self, other_series: "GridDataSeries") -> "MultiGridDataSeries":
        """GridDataSeriesを結合する.

        Parameters
        ----------
        other_series : GridDataSeries
            結合するGridDataSeries

        Returns
        -------
        MultiGridDataSeries
            結合したGridDataSeries
        """
        return MultiGridDataSeries(self, other_series)

    def __add__(self, other_series: "GridDataSeries") -> "MultiGridDataSeries":
        """GridDataSeriesを結合する.

        Parameters
        ----------
        other_series : GridDataSeries
            結合するGridDataSeries

        Returns
        -------
        MultiGridDataSeries
            結合したGridDataSeries
        """
        if not isinstance(other_series, GridDataSeries):
            raise TypeError()

        return self.chain(other_series)

    def __iter__(self):
        indexes = sorted(self._index2key.keys())
        for index in indexes:
            yield self[index]

    def __len__(self):
        return len(self._index2key)


class MultiGridDataSeries(GridDataSeries):
    """連続する複数の3次元時系列データを管理する.

    Attributes
    ----------
    datafile : DataFileInfo
        データファイル情報
    name : str
        データセット名
    tunit : UnitTranslator
        時間の単位変換器
    axisunit : UnitTranslator
        空間軸の単位変換器
    valunit : UnitTranslator
        値の単位変換器
    """

    def __init__(self, *series):
        self.series = []
        for data in series:
            self.series += self.__expand(data)

        self.datafile = self.series[0].datafile
        self.tunit = self.series[0].tunit
        self.axisunit = self.series[0].axisunit
        self.valunit = self.series[0].valunit

        self.name = self.series[0].name

    def __expand(
        self, data_series: Union["GridDataSeries", "MultiGridDataSeries"]
    ) -> List[GridDataSeries]:
        """与えられたオブジェクトがMultiGridDataSeriesなら展開してGridDataSeriesのリストとして返す.

        Parameters
        ----------
        data_series : GridDataSeries or MultiGridDataSeries
            オブジェクト

        Returns
        -------
        list(GridDataSeries)
            GridDataSeriesのリスト

        Raises
        ------
        TypeError
            オブジェクトがGridDataSeriesでない場合の例外
        """
        if not isinstance(data_series, GridDataSeries):
            raise TypeError()
        if not isinstance(data_series, MultiGridDataSeries):
            return [data_series]

        # data_seriesがMultiGridDataSeriesならデータを展開して結合する.
        expanded = []
        for data in data_series.series:
            expanded += self.__expand(data)

        return expanded

    def close(self) -> None:
        """hdf5ファイルを閉じる."""
        for data in self.series:
            self.series.h5.close()

    def time_series(
        self, x: Union[int, slice], y: Union[int, slice], z: Union[int, slice]
    ):
        """指定した範囲の時系列データを取得する.

        Parameters
        ----------
        x : int or slice
            x座標
        y : int or slice
            y座標
        z : int or slice
            z座標

        Returns
        -------
        numpy.ndarray
            指定した範囲の時系列データ
        """
        series = np.concatenate([data.time_series(x, y, z) for data in self.series])
        return series

    @property
    def filename(self) -> Path:
        """先頭データのファイル名を返す.

        Returns
        -------
        Path
            ファイル名
        """
        return self.series[0].datafile.filename

    @property
    def filenames(self) -> List[Path]:
        """ファイル名のリストを返す.

        Returns
        -------
        list(Path)
            ファイル名のリスト
        """
        return [data.filename for data in self.series]

    @property
    def directory(self) -> Path:
        """先頭データのディレクトリ名を返す.

        Returns
        -------
        Path
            ディレクトリ名
        """
        return self.series[0].datafile.directory

    @property
    def directories(self) -> List[Path]:
        """ディレクトリ名のリストを返す.

        Returns
        -------
        list(Path)
            ディレクトリ名のリスト
        """
        return [data.directory for data in self.series]

    def _create_data_with_index(self, index: int) -> Data3d:
        """時間が指定された場合に、その時間におけるData3dを生成する.

        Parameters
        ----------
        index : int
            時間インデックス

        Returns
        -------
        Data3d
            生成したData3d

        Raises
        ------
        IndexError
            指定した時間が存在しない場合の例外
        """
        if index < len(self.series[0]):
            return self.series[0][index]

        length = len(self.series[0])
        for series in self.series[1:]:
            # 先頭データは前のデータの最後尾と重複しているためカウントしない
            series_len = len(series) - 1

            if index < series_len + length:
                local_index = index - length + 1
                return series[local_index]

            length += series_len

        raise IndexError()

    def __iter__(self):
        iters = [iter(self.series[0])]
        for data in self.series[1:]:
            it = iter(data)
            next(it)  # 先頭データを捨てる
            iters.append(it)
        return chain(iters)

    def __len__(self) -> int:
        # 先頭データは前のデータの最後尾と重複しているためカウントしない
        return np.sum([len(data) for data in self.series]) - (len(self.series) - 1)



class UnitTranslator:
    """単位変換器.

    Attributes
    ----------
    from_unit : float
        変換前の値
    to_unit : float
        変換後の値
    ratio : float
        変換係数 (変換後 = 変換係数 * 変換前)
    name : str or None
        単位の名前(例: "Mass", "Frequency")
    unit : str or None
        単位(例: "kg", "Hz")
    """

    def __init__(self, from_unit, to_unit, name=None, unit=None):
        """単位変換器を生成する.

        Parameters
        ----------
        from_unit : float
            変換前の値
        to_unit : float
            変換後の値
        name : str or None
            単位の名前(例: "Mass", "Frequency"), by default None
        unit : str or None
            単位(例: "kg", "Hz"), by default None
        """
        self.from_unit = from_unit
        self.to_unit = to_unit
        self.ratio = to_unit / from_unit
        self.name = name
        self.unit = unit

    def set_name(self, name, unit=None):
        """名前を設定する.

        Parameters
        ----------
        name : str
            名前
        unit : str
            単位

        Returns
        -------
        UnitTranslator
            self
        """
        self.name = name
        self.unit = unit
        return self

    def trans(self, value, reverse=False):
        """単位変換を行う.

        Parameters
        ----------
        value : float
            変換前の値(reverse=Trueの場合変換後の値)
        reverse : bool, optional
            逆変換を行う場合True, by default False

        Returns
        -------
        float
            変換後の値(reverse=Trueの場合変換前の値)
        """
        if reverse:
            return value / self.ratio
        else:
            return value * self.ratio

    def reverse(self, value):
        """単位逆変換を行う.

        Parameters
        ----------
        value : float
            変換後の値

        Returns
        -------
        float
            変換前の値
        """
        return self.trans(value, reverse=True)

    def __mul__(self, other):
        from_unit = self.from_unit * other.from_unit
        to_unit = self.to_unit * other.to_unit
        return UnitTranslator(from_unit, to_unit)

    def __rmul__(self, other):
        other = UnitTranslator(other, other)
        return other * self

    def __truediv__(self, other):
        from_unit = self.from_unit / other.from_unit
        to_unit = self.to_unit / other.to_unit
        return UnitTranslator(from_unit, to_unit)

    def __rtruediv__(self, other):
        other = UnitTranslator(other, other)
        return other / self

    def __pow__(self, other):
        from_unit = self.from_unit**other
        to_unit = self.to_unit**other
        return UnitTranslator(from_unit, to_unit)

    def __str__(self):
        return "{}({:.4}, [{}])".format(self.name, self.ratio, self.unit)

    def __repr__(self):
        return self.__str__()


class Units:
    """EMSES用の単位変換器を管理する.

    SI単位系からEMSES単位系への変換を行う.
    """

    def __init__(self, dx: float, to_c: float):
        """EMSES用の単位変換器を生成する.

        Parameters
        ----------
        dx : float, optional
            Grid length [m]
        to_c : float
            Light speed in EMSES unit
        """
        self.dx = dx
        """Grid length [m]"""
        self.to_c = to_c
        """Light speed in EMSES"""

        from_c = 299792458
        to_e0 = 1
        pi = UnitTranslator(3.141592654, 3.141592654)
        e = UnitTranslator(2.718281828, 2.718281828)

        c = UnitTranslator(from_c, to_c)
        v = 1 * c

        _m0 = 4 * pi.from_unit * 1e-7
        e0 = UnitTranslator(1 / (_m0 * c.from_unit**2), to_e0)
        eps = 1 * e0
        mu = 1 / eps / v**2
        m0 = UnitTranslator(_m0, mu.trans(_m0))

        kB = UnitTranslator(1.38065052e-23, 1.38065052e-23)

        length = UnitTranslator(dx, 1)
        t = length / v
        f = 1 / t
        n = 1 / (length**3)
        N = v * n

        _qe = 1.6021765e-19
        _me = 9.1093819e-31
        _mi = 1.67261e-27
        qe_me = UnitTranslator(-_qe / _me, -1)
        q_m = 1 * qe_me

        q = e0 / q_m * length * v**2
        m = q / q_m

        qe = UnitTranslator(_qe, q.trans(_qe))
        me = UnitTranslator(_me, m.trans(_me))
        mi = UnitTranslator(_mi, m.trans(_mi))
        rho = q / length**3

        F = m * length / t**2
        P = F * v
        W = F * length
        w = W / (length**3)

        i = q / length * v
        J = i / length**2
        phi = v**2 / q_m
        E = phi / length
        H = i / length
        C = eps * length
        R = phi / i
        G = 1 / R

        B = v / length / q_m
        L = mu * length
        T = W / kB

        a = v / t

        EC = G / length

        self.pi = pi.set_name("Circular constant", unit="")
        """Unit translator for Circular constant []"""

        self.e = e.set_name("Napiers constant", unit="")
        """Unit translator for Napiers constant []"""

        self.c = c.set_name("Light Speed", unit="m/s")
        """Unit translator for Light Speed [m/s]"""

        self.e0 = e0.set_name("FS-Permttivity", unit="F/m")
        """Unit translator for FS-Permttivity [F/m]"""

        self.m0 = m0.set_name("FS-Permeablity", unit="N/A^2")
        """Unit translator for FS-Permeablity [N/A^2]"""

        self.qe = qe.set_name("Elementary charge", unit="C")
        """Unit translator for Elementary charge [C]"""

        self.me = me.set_name("Electron mass", unit="kg")
        """Unit translator for Electron mass [kg]"""

        self.mi = mi.set_name("Proton mass", unit="kg")
        """Unit translator for Proton mass [kg]"""

        self.qe_me = qe_me.set_name("Electron charge-to-mass ratio", unit="C/kg")
        """Unit translator for Electron charge-to-mass ratio [C/kg]"""

        self.kB = kB.set_name("Boltzmann constant", unit="J/K")
        """Unit translator for Boltzmann constant [J/K]"""

        self.length = length.set_name("Sim-to-Real length ratio", unit="m")
        """Unit translator for Sim-to-Real length ratio [m]"""

        self.m = m.set_name("Mass", unit="kg")
        """Unit translator for Mass [kg]"""

        self.t = t.set_name("Time", unit="s")
        """Unit translator for Time [s]"""

        self.f = f.set_name("Frequency", unit="Hz")
        """Unit translator for Frequency [Hz]"""

        self.v = v.set_name("Velocity", unit="m/s")
        """Unit translator for Velocity [m/s]"""

        self.n = n.set_name("Number density", unit="/m^3")
        """Unit translator for Number density [/m^3]"""

        self.N = N.set_name("Flux", unit="/m^2s")
        """Unit translator for Flux [/m^2s]"""

        self.F = F.set_name("Force", unit="N")
        """Unit translator for Force [N]"""

        self.P = P.set_name("Power", unit="W")
        """Unit translator for Power [W]"""

        self.W = W.set_name("Energy", unit="J")
        """Unit translator for Energy [J]"""

        self.w = w.set_name("Energy density", unit="J/m^3")
        """Unit translator for Energy density [J/m^3]"""

        self.eps = eps.set_name("Permittivity", unit="F/m")
        """Unit translator for Permittivity  [F/m]"""

        self.q = q.set_name("Charge", unit="C")
        """Unit translator for Charge [C]"""

        self.rho = rho.set_name("Charge density", unit="C/m^3")
        """Unit translator for Charge density [C/m^3]"""

        self.q_m = q_m.set_name("Charge-to-mass ratio", unit="C/kg")
        """Unit translator for Charge-to-mass ratio [C/kg]"""

        self.i = i.set_name("Current", unit="A")
        """Unit translator for Current [A]"""

        self.J = J.set_name("Current density", unit="A/m^2")
        """Unit translator for Current density [A/m^2]"""

        self.phi = phi.set_name("Potential", unit="V")
        """Unit translator for Potential [V]"""

        self.E = E.set_name("Electric field", unit="V/m")
        """Unit translator for Electric field [V/m]"""

        self.H = H.set_name("Magnetic field", unit="A/m")
        """Unit translator for Magnetic field [A/m]"""

        self.C = C.set_name("Capacitance", unit="F")
        """Unit translator for Capacitance [F]"""

        self.R = R.set_name("Resistance", unit="Ω")
        """Unit translator for Resistance [Ω]"""

        self.G = G.set_name("Conductance", unit="S")
        """Unit translator for Conductance [S]"""

        self.mu = mu.set_name("Permiability", unit="H/m")
        """Unit translator for Permiability [H/m]"""

        self.B = B.set_name("Magnetic flux density", unit="T")
        """Unit translator for Magnetic flux density [T]"""

        self.L = L.set_name("Inductance", unit="H")
        """Unit translator for Inductance [H]"""

        self.T = T.set_name("Temperature", unit="K")
        """Unit translator for Temperature [K]"""

        self.a = a.set_name("Acceleration", unit="m/s^2")
        """Unit translator for Acceleration [m/s^2]"""

        self.EC = EC.set_name("Electric conductivity", unit="S/m")
        """Unit translator for Electric conductivity [S/m]"""

    def translators(self):
        """変換器のリストを返す.

        Returns
        -------
        list(UnitTranslator)
            変換器のリスト
        """
        return
        # translators = inspect.getmembers(
        #     self, lambda x: isinstance(x, UnitTranslator))
        # return list(map(lambda x: x[1], translators))



from typing import Tuple, Union

import matplotlib.pyplot as plt
import numpy as np

from emout import Emout
from emout.utils.emsesinp import InpFile
from emout.utils.units import Units


def plot_surface_with_hole(
    data_xyz, inp, add_colorbar=True, show=False, vrange="minmax", **kwargs
):
    xl = int(inp.xlrechole[1])
    xu = int(inp.xurechole[1])
    yl = int(inp.ylrechole[1])
    yu = int(inp.yurechole[1])
    zl = int(inp.zlrechole[1])
    zu = int(inp.zurechole[0])

    if kwargs.get("use_si", False):
        vmax = data_xyz.val_si.max()
        vmin = data_xyz.val_si.min()
    else:
        vmax = data_xyz.max()
        vmin = data_xyz.min()

    if vrange == "minmax":
        pass
    elif vrange == "center":
        vmax = max(abs(vmax), abs(vmin))
        vmin = -vmax

    kwargs["vmax"] = kwargs.get("vmax", None) or vmax
    kwargs["vmin"] = kwargs.get("vmin", None) or vmin

    # plot top surface
    data_xyz[zu, : yl + 1, :].plot(mode="surf", **kwargs)
    data_xyz[zu, yu:, :].plot(mode="surf", **kwargs)
    data_xyz[zu, yl : yu + 1, : xl + 1].plot(mode="surf", **kwargs)
    data_xyz[zu, yl : yu + 1, xu:].plot(mode="surf", **kwargs)

    # plot wall
    data_xyz[zl : zu + 1, yl, xl : xu + 1].plot(mode="surf", **kwargs)
    data_xyz[zl : zu + 1, yu, xl : xu + 1].plot(mode="surf", **kwargs)
    data_xyz[zl : zu + 1, yl : yu + 1, xl].plot(mode="surf", **kwargs)
    data_xyz[zl : zu + 1, yl : yu + 1, xu].plot(mode="surf", **kwargs)

    # plot bottom
    data_xyz[zl, yl : yu + 1, xl : xu + 1].plot(
        mode="surf", add_colorbar=add_colorbar, **kwargs
    )

    if show:
        plt.show()


def plot_hole_line(
    inp_or_data: Union[InpFile, Emout],
    unit: Units = None,
    use_si: bool = True,
    offsets: Tuple[int, int] = (0, 0),
    axis="xz",
    color="black",
    linewidth=None,
    fix_lims=True,
):
    if isinstance(inp_or_data, InpFile):
        inp: InpFile = inp_or_data
    else:
        data: Emout = inp_or_data
        inp = data.inp
        unit = data.unit

    if fix_lims:
        plt.xlim(plt.xlim())
        plt.ylim(plt.ylim())

    xl = inp.xlrechole[1]
    xu = inp.xurechole[1]
    yl = inp.ylrechole[1]
    yu = inp.yurechole[1]
    zl = inp.zlrechole[1] + 0.5
    zu = inp.zurechole[0]
    nx = inp.nx
    ny = inp.ny
    nz = inp.nz

    if axis == "xz":
        xs = np.array([0.0, xl, xl, xu, xu, nx - 1])
        ys = np.array([zu, zu, zl, zl, zu, zu])

    if use_si and unit is not None:
        xs = unit.length.reverse(xs)
        ys = unit.length.reverse(ys)

    im = plt.plot(xs + offsets[0], ys + offsets[1], color=color, linewidth=linewidth)

    return im


def plot_line_of_hole_half(inp, off, unit):
    xl = int(inp.xlrechole[1])
    xu = int(inp.xurechole[1])
    yl = int(inp.ylrechole[1])
    yu = int(inp.yurechole[1])
    zl = int(inp.zlrechole[1])
    zu = int(inp.zurechole[0])

    xc = (xl + xu) // 2
    yc = (yl + yu) // 2

    xc_si = unit.reverse(xc)
    yc_si = unit.reverse(yc)
    zu_si = unit.reverse(zu)

    ax = plt.gca()

    surf_points = np.array(
        [
            [xc, yl - off, zu],
            [xc, yl, zu],
            [xl, yl, zu],
            [xl, yu, zu],
            [xc, yu, zu],
            [xc, yu + off, zu],
            [xc - off, yu + off, zu],
            [xc - off, yl - off, zu],
            [xc, yl - off, zu],
        ]
    )

    bottom_points = np.array(
        [[xc, yl, zl], [xc, yu, zl], [xl, yu, zl], [xl, yl, zl], [xc, yl, zl]]
    )

    bottom2_points = np.array(
        [
            [xc, yl - off, zl - off // 2],
            [xc, yu + off, zl - off // 2],
            [xc - off, yu + off, zl - off // 2],
            [xc - off, yl - off, zl - off // 2],
            [xc, yl - off, zl - off // 2],
        ]
    )

    lines = np.array(
        [
            [[xc, yl, zl], [xc, yl, zu]],
            [[xl, yl, zl], [xl, yl, zu]],
            [[xl, yu, zl], [xl, yu, zu]],
            [[xc, yu, zl], [xc, yu, zu]],
            [[xc, yl - off, zl - off // 2], [xc, yl - off, zu]],
            [[xc, yu + off, zl - off // 2], [xc, yu + off, zu]],
            [[xc - off, yu + off, zl - off // 2], [xc - off, yu + off, zu]],
            [[xc - off, yl - off, zl - off // 2], [xc - off, yl - off, zu]],
        ]
    )

    surf_points = unit.reverse(surf_points)
    bottom_points = unit.reverse(bottom_points)
    bottom2_points = unit.reverse(bottom2_points)
    lines = unit.reverse(lines)

    def parse(points, offsets):
        x = points[:, 0] + offsets[0]
        y = points[:, 1] + offsets[1]
        z = points[:, 2] + offsets[2]
        return x, y, z

    offsets = (-xc_si, -yc_si, -zu_si)
    ax.plot(*parse(surf_points, offsets), color="black")
    ax.plot(*parse(bottom_points, offsets), color="black")
    ax.plot(*parse(bottom2_points, offsets), color="black")

    for line in lines:
        ax.plot(*parse(line, offsets), color="black")


def plot_surface_with_hole_half(
    data_xyz, inp, off=10, add_colorbar=True, show=False, vrange="minmax", **kwargs
):
    xl = int(inp.xlrechole[1])
    xu = int(inp.xurechole[1])
    yl = int(inp.ylrechole[1])
    yu = int(inp.yurechole[1])
    zl = int(inp.zlrechole[1])
    zu = int(inp.zurechole[0])

    xlen = (xu - xl) / 2 + off
    ylen = yu - yl + off * 2
    zlen = zu - zl + off

    lenmax = max(max(xlen, ylen), zlen)
    box_aspect = (xlen / lenmax, ylen / lenmax, zlen / lenmax)

    plt.gca(projection="3d").set_box_aspect(box_aspect)

    xc = (xl + xu) // 2
    yc = (xl + xu) // 2
    zc = (xl + xu) // 2

    if kwargs.get("use_si", False):
        vmax = data_xyz.val_si.max()
        vmin = data_xyz.val_si.min()
    else:
        vmax = data_xyz.max()
        vmin = data_xyz.min()

    if vrange == "minmax":
        pass
    elif vrange == "center":
        vmax = max(abs(vmax), abs(vmin))
        vmin = -vmax

    kwargs["vmax"] = kwargs.get("vmax", None) or vmax
    kwargs["vmin"] = kwargs.get("vmin", None) or vmin

    # plot top surface
    data_xyz[zu, yl - off : yl + 1, xl - off : xc + 1].plot(mode="surf", **kwargs)
    data_xyz[zu, yu : yu + off, xl - off : xc + 1].plot(mode="surf", **kwargs)
    data_xyz[zu, yl : yu + 1, xl - off : xl + 1].plot(mode="surf", **kwargs)

    # plot wall
    data_xyz[zl : zu + 1, yl, xl : xc + 1].plot(mode="surf", **kwargs)
    data_xyz[zl : zu + 1, yu, xl : xc + 1].plot(mode="surf", **kwargs)
    data_xyz[zl : zu + 1, yl : yu + 1, xl].plot(mode="surf", **kwargs)

    # innner wall
    def alltrue(x):
        return x == x

    masked = data_xyz.masked(alltrue)

    masked[zl - off // 2 : zu + 1, yl - off : yl + 1, xc].plot(mode="surf", **kwargs)
    masked[zl - off // 2 : zu + 1, yu : yu + off, xc].plot(mode="surf", **kwargs)
    masked[zl - off // 2 : zu + 1, yl - off, xl - off : xc + 1].plot(
        mode="surf", **kwargs
    )
    masked[zl - off // 2 : zu + 1, yu + off - 1, xl - off : xc + 1].plot(
        mode="surf", **kwargs
    )
    masked[zl - off // 2 : zl + 1, yl : yu + 1, xc].plot(mode="surf", **kwargs)

    # plot bottom
    data_xyz[zl, yl : yu + 1, xl : xc + 1].plot(
        mode="surf",
        add_colorbar=add_colorbar,
        xlabel="x[m]",
        ylabel="y[m]",
        zlabel="z[m]",
        **kwargs
    )

    if show:
        plt.show()



import re
import warnings
from os import PathLike
from typing import Any, List, Literal, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np

import emout.plot.basic_plot as emplt
import emout.utils as utils
from emout.plot.animation_plot import ANIMATER_PLOT_MODE, FrameUpdater
from emout.utils import UnitTranslator


class VectorData(utils.Group):
    def __init__(self, objs: List[Any], name=None, attrs=None):
        x_data, y_data = objs

        if attrs is None:
            attrs = dict()

        if name:
            attrs["name"] = name
        elif "name" in attrs:
            pass
        elif hasattr(x_data, "name"):
            attrs["name"] = name
        else:
            attrs["name"] = ""

        super().__init__([x_data, y_data], attrs=attrs)
        self.x_data = x_data
        self.y_data = y_data

    def __setattr__(self, key, value):
        if key in ("x_data", "y_data"):
            super().__dict__[key] = value
            return
        super().__setattr__(key, value)

    @property
    def name(self) -> str:
        return self.attrs["name"]

    @property
    def valunit(self) -> UnitTranslator:
        return self.objs[0].valunit

    @property
    def axisunits(self) -> UnitTranslator:
        return self.objs[0].axisunits

    @property
    def slice_axes(self) -> np.ndarray:
        return self.objs[0].slice_axes

    @property
    def slices(self) -> np.ndarray:
        return self.objs[0].slices

    @property
    def shape(self) -> np.ndarray:
        return self.objs[0].shape

    def build_frame_updater(
        self,
        axis: int = 0,
        title: Union[str, None] = None,
        notitle: bool = False,
        offsets: Union[
            Tuple[Union[float, str], Union[float, str], Union[float, str]], None
        ] = None,
        use_si: bool = True,
        vmin: float = None,
        vmax: float = None,
        **kwargs,
    ):
        """アニメーション描画処理を構築する.

        Parameters
        ----------
        axis : int, optional
            アニメーションする軸, by default 0
        title : str, optional
            タイトル(Noneの場合データ名(phisp等)), by default None
        notitle : bool, optional
            タイトルを付けない場合True, by default False
        offsets : (float or str, float or str, float or str)
            プロットのx,y,z軸のオフセット('left': 最初を0, 'center': 中心を0, 'right': 最後尾を0, float: 値だけずらす), by default None
        use_si : bool
            SI単位系を用いる場合True(そうでない場合EMSES単位系を用いる), by default False
        vmin : float, optional
            最小値, by default None
        vmax : float, optional
            最大値, by default None
        """
        if vmin is None:
            vmin = min(self.objs[0].min(), self.objs[1].min())
        if vmax is None:
            vmax = max(self.objs[0].max(), self.objs[1].max())

        updater = FrameUpdater(
            self, axis, title, notitle, offsets, use_si, vmin, vmax, **kwargs
        )

        return updater

    def gifplot(
        self,
        fig: Union[plt.Figure, None] = None,
        axis: int = 0,
        action: ANIMATER_PLOT_MODE = "to_html",
        filename: PathLike = None,
        interval: int = 200,
        repeat: bool = True,
        title: Union[str, None] = None,
        notitle: bool = False,
        offsets: Union[
            Tuple[Union[float, str], Union[float, str], Union[float, str]], None
        ] = None,
        use_si: bool = True,
        vmin: float = None,
        vmax: float = None,
        show: bool = False,
        savefilename: PathLike = None,
        to_html: bool = False,
        return_updater: bool = False,
        **kwargs,
    ):
        """gifアニメーションを作成する

        Parameters
        ----------
        fig : Figure
            アニメーションを描画するFigure(Noneの場合新しく作成する), by default None
        axis : int, optional
            アニメーションする軸, by default 0

        action : {'return', 'show', 'to_html', 'save', 'frames'}, optional, by default 'to_html'
            Determines the behavior of the function:

            - 'return': The plot object is returned without rendering it.
            - 'show': The plot is displayed immediately.
            - 'to_html': The plot is converted to an Ipython.display.HTML object and returned.
            - 'save': The plot is saved to a file specified by 'filename' argument.
            - 'frames': FrameUpdater object is returned without rendering it.

        filename : str, optional
            保存するファイル名(actionが'save'以外の場合やNoneの場合保存されない), by default None

        interval : int, optional
            フレーム間のインターバル(ミリ秒), by default 400
        repeat : bool
            アニメーションをループするならTrue, by default True
        title : str, optional
            タイトル(Noneの場合データ名(phisp等)), by default None
        notitle : bool, optional
            タイトルを付けない場合True, by default False
        offsets : (float or str, float or str, float or str)
            プロットのx,y,z軸のオフセット('left': 最初を0, 'center': 中心を0, 'right': 最後尾を0, float: 値だけずらす), by default None
        use_si : bool
            SI単位系を用いる場合True(そうでない場合EMSES単位系を用いる), by default False

        show : bool, optional
            プロットを表示する場合True(ファイルに保存する場合は非表示), by default

            .. deprecated :: 1.2.1

               This parameter is deprecated and will be removed in version 2.0.0.
               Use the 'action'='show' instead for equivalent functionality.

        savefilename : str, optional
            保存するファイル名(Noneの場合保存しない), by default None

            .. deprecated :: 1.2.1

               This parameter is deprecated and will be removed in version 2.0.0.
               Use the plot('action'='save', filename='example.gif') instead for equivalent functionality.

        to_html : bool
            アニメーションをHTMLとして返す. (使用例: Jupyter Notebook等でアニメーションを描画する際等)

            .. deprecated :: 1.2.1

               This parameter is deprecated and will be removed in version 2.0.0.
               Use the 'action'='to_html' instead for equivalent functionality.

        return_updater : bool
            FrameUpdaterを返す場合True, by default False

            .. deprecated :: 1.2.1

               This parameter is deprecated and will be removed in version 2.0.0.
               Use the 'action'='frames' instead for equivalent functionality.

        Returns
        -------
        Depending on the selected action:

        - If 'return': Returns the tuple of the plot object (fig, animation).
        - If 'show': Does not return anything, displays the plot.
        - If 'to_html': Returns an Ipython.display.HTML object of the plot (for Jupyter).
        - If 'save': Does not return anything, saves the plot to a file.
        - If 'frames': Returns FrameUpdater object.

        Examples
        --------
        >>> fig, ani = gifplot(action="return")
        Returns the tuple of the plot object.

        >>> gifplot(action="show")
        Displays the plot.

        >>> html = gifplot(action="to_html")
        Returns the HTML representation of the plot.

        >>> gifplot(action="save", filename = "example.gif")
        Saves the plot to a file.

        >>> updater = gifplot(action="frames")
        Returns FrameUpdater object.
        """
        if return_updater:
            warnings.warn(
                "The 'return_updater' flag is deprecated. "
                "Please use gifplot(action='frames') instead.",
                DeprecationWarning,
            )
            action = "frames"

        print(f"vector: {action}")

        updater = self.build_frame_updater(
            axis, title, notitle, offsets, use_si, vmin, vmax, **kwargs
        )

        if action == "frames":
            return updater

        animator = updater.to_animator([[[updater]]])

        return animator.plot(
            fig=fig,
            action=action,
            filename=filename,
            show=show,
            savefilename=savefilename,
            interval=interval,
            repeat=repeat,
            to_html=to_html,
            **kwargs,
        )

    def plot(
        self,
        *args,
        **kwargs,
    ):
        if self.x_data.ndim == 2:
            self.plot2d(
                *args,
                **kwargs,
            )

    def plot2d(
        self,
        mode: Literal["stream", "vec"] = "stream",
        axes: Literal["auto", "xy", "yz", "zx", "yx", "zy", "xy"] = "auto",
        show: bool = False,
        use_si: bool = True,
        offsets: Union[
            Tuple[Union[float, str], Union[float, str], Union[float, str]], None
        ] = None,
        **kwargs,
    ):
        """2次元データをプロットする.

        Parameters
        ----------
        mode : str
            プロットの種類('vec': quiver plot, 'stream': streamline plot), by default 'stream'
        axes : str, optional
            プロットする軸('xy', 'zx', etc), by default 'auto'
        show : bool
            プロットを表示する場合True(ファイルに保存する場合は非表示), by default False
        use_si : bool
            SI単位系を用いる場合True(そうでない場合EMSES単位系を用いる), by default False
        offsets : (float or str, float or str, float or str)
            プロットのx,y,z軸のオフセット('left': 最初を0, 'center': 中心を0, 'right': 最後尾を0, float: 値だけずらす), by default None
        mesh : (numpy.ndarray, numpy.ndarray), optional
            メッシュ, by default None
        savefilename : str, optional
            保存するファイル名(Noneの場合保存しない), by default None
        cmap : matplotlib.Colormap or str or None, optional
            カラーマップ, by default cm.coolwarm
        vmin : float, optional
            最小値, by default None
        vmax : float, optional
            最大値, by default None
        figsize : (float, float), optional
            図のサイズ, by default None
        xlabel : str, optional
            x軸のラベル, by default None
        ylabel : str, optional
            y軸のラベル, by default None
        title : str, optional
            タイトル, by default None
        interpolation : str, optional
            用いる補間方法, by default 'bilinear'
        dpi : int, optional
            解像度(figsizeが指定された場合は無視される), by default 10

        Returns
        -------
        AxesImage or None
            プロットしたimageデータ(保存またはshowした場合None)

        Raises
        ------
        Exception
            プロットする軸のパラメータが間違っている場合の例外
        Exception
            プロットする軸がデータにない場合の例外
        Exception
            データの次元が2でない場合の例外
        """
        if self.objs[0].valunit is None:
            use_si = False

        if axes == "auto":
            axes = "".join(sorted(self.objs[0].use_axes))

        if not re.match(r"x[yzt]|y[xzt]|z[xyt]|t[xyz]", axes):
            raise Exception(
                'Error: axes "{axes}" cannot be used with Data2d'.format(axes=axes)
            )
        if axes[0] not in self.objs[0].use_axes or axes[1] not in self.objs[0].use_axes:
            raise Exception(
                'Error: axes "{axes}" cannot be used because {axes}-axis does not exist in this data.'.format(
                    axes=axes
                )
            )
        if len(self.objs[0].shape) != 2:
            raise Exception(
                'Error: axes "{axes}" cannot be used because data is not 2dim shape.'.format(
                    axes=axes
                )
            )

        # x: 3, y: 2, z:1 t:0
        axis1 = self.objs[0].slice_axes[self.objs[0].use_axes.index(axes[0])]
        axis2 = self.objs[0].slice_axes[self.objs[0].use_axes.index(axes[1])]

        x = np.arange(*utils.slice2tuple(self.objs[0].slices[axis1]), dtype=float)
        y = np.arange(*utils.slice2tuple(self.objs[0].slices[axis2]), dtype=float)

        if use_si:
            xunit = self.objs[0].axisunits[axis1]
            yunit = self.objs[0].axisunits[axis2]
            valunit = self.objs[0].valunit

            x = xunit.reverse(x)
            y = yunit.reverse(y)

            _xlabel = "{} [{}]".format(axes[0], xunit.unit)
            _ylabel = "{} [{}]".format(axes[1], yunit.unit)
            _title = "{} [{}]".format(self.name, valunit.unit)

            x_data = self.x_data.val_si
            y_data = self.y_data.val_si
        else:
            _xlabel = axes[0]
            _ylabel = axes[1]
            _title = self.name

            x_data = self.x_data
            y_data = self.y_data

        def _offseted(line, offset):
            line = line.astype(float)
            if offset == "left":
                line -= line[0]
            elif offset == "center":
                line -= line[len(line) // 2]
            elif offset == "right":
                line -= line[-1]
            else:
                line += offset
            return line

        if offsets is not None:
            x = _offseted(x, offsets[0])
            y = _offseted(y, offsets[1])

        kwargs["xlabel"] = kwargs.get("xlabel", None) or _xlabel
        kwargs["ylabel"] = kwargs.get("ylabel", None) or _ylabel
        kwargs["title"] = kwargs.get("title", None) or _title

        mesh = np.meshgrid(x, y)
        if "vec" in mode:
            img = emplt.plot_2d_vector(x_data, y_data, mesh=mesh, **kwargs)
        elif "stream" in mode:
            img = emplt.plot_2d_streamline(x_data, y_data, mesh=mesh, **kwargs)

        if show:
            plt.show()
            return None
        else:
            return img


VectorData2d = VectorData



import logging
import re
import warnings
from pathlib import Path
from typing import Callable, Literal, Union

import h5py
import numpy as np
import pandas as pd
import scipy.constants as cn
from tqdm import tqdm
from tqdm.notebook import tqdm as notebook_tqdm

from emout.utils import InpFile, RegexDict, UnitConversionKey, Units, UnitTranslator

from .griddata_series import GridDataSeries
from .vector_data import VectorData2d

logger = logging.getLogger(__name__)


def get_tqdm():
    """Function to determine the appropriate tqdm version."""
    try:
        # Check if the environment is Jupyter Notebook
        shell = get_ipython().__class__.__name__  # type: ignore
        if shell == "ZMQInteractiveShell":  # Jupyter Notebook environment
            logger.debug("Jupyter Notebook environment detected.")
            return notebook_tqdm
        else:  # IPython environment, but not Jupyter Notebook
            logger.debug("IPython environment detected (not Jupyter).")
            return tqdm
    except NameError:  # IPython is not available = standard Python environment
        logger.debug("Standard Python environment detected.")
        return tqdm


# Get the appropriate tqdm
tqdm = get_tqdm()


def t_unit(out: "Emout") -> UnitTranslator:
    """tの単位変換器を生成する.

    Parameters
    ----------
    out : Emout
        Emoutオブジェクト

    Returns
    -------
    UnitTranslator
        tの単位変換器
    """
    return (out.unit.t * UnitTranslator(out.inp.ifdiag * out.inp.dt, 1)).set_name(
        "t", unit="s"
    )


def wpet_unit(out: "Emout") -> UnitTranslator:
    """wpe * tの単位変換器を生成する.

    以下のコードを実行することで、データのt軸をwpe*tで規格化できる.

    >>> Emout.name2unit['t'] = wpet_unit

    Parameters
    ----------
    out : Emout
        Emoutオブジェクト

    Returns
    -------
    UnitTranslator
        wpe * tの単位変換器
    """
    return UnitTranslator(
        out.inp.wp[0] * out.inp.ifdiag * out.inp.dt, 1, name="wpe * t", unit=""
    )


def wpit_unit(out: "Emout") -> UnitTranslator:
    """wpi * tの単位変換器を生成する.

    以下のコードを実行することで、データのt軸をwpe*tで規格化できる.

    >>> Emout.name2unit['t'] = wpit_unit

    Parameters
    ----------
    out : Emout
        Emoutオブジェクト

    Returns
    -------
    UnitTranslator
        wpi * tの単位変換器
    """
    return UnitTranslator(
        out.inp.wp[1] * out.inp.ifdiag * out.inp.dt, 1, name="wpi * t", unit=""
    )


def none_unit(out: "Emout") -> UnitTranslator:
    return UnitTranslator(1, 1, name="", unit="")


def ndp_unit(ispec: int) -> Callable[["Emout"], UnitTranslator]:
    def ndp_unit(out: "Emout") -> UnitTranslator:
        wp = out.unit.f.reverse(out.inp.wp[ispec])
        mp = abs(cn.m_e / out.inp.qm[ispec])
        np = wp**2 * mp * cn.epsilon_0 / cn.e**2
        return UnitTranslator(np * 1e-6, 1.0, name="number density", unit="/cc")

    return ndp_unit


def nd3p_unit(out: "Emout") -> UnitTranslator:
    wpp = out.unit.f.reverse(out.inp.wp[2])
    np = wpp**2 * cn.m_e * cn.epsilon_0 / cn.e**2
    return UnitTranslator(np * 1e-6, 1.0, name="number density", unit="/cc")


class Emout:
    """EMSES出力・inpファイルを管理する.

    Attributes
    ----------
    directory : Path
        管理するディレクトリ
    dataname : GridData
        3次元データ(datanameは"phisp"などのhdf5ファイルの先頭の名前)
    """

    name2unit = RegexDict(
        {
            r"phisp": lambda self: self.unit.phi,
            # r'nd[12]p': ndp_unit,
            r"nd1p": ndp_unit(0),
            r"nd2p": ndp_unit(1),
            r"nd3p": ndp_unit(2),
            r"nd4p": ndp_unit(3),
            r"nd5p": ndp_unit(4),
            r"nd6p": ndp_unit(5),
            r"nd7p": ndp_unit(6),
            r"nd8p": ndp_unit(7),
            r"nd9p": ndp_unit(8),
            r"rho": lambda self: self.unit.rho,
            r"rhobk": lambda self: self.unit.rho,
            r"j.*": lambda self: self.unit.J,
            r"b[xyz]": lambda self: self.unit.H,
            r"rb[xyz]": lambda self: self.unit.H,
            r"e[xyz]": lambda self: self.unit.E,
            r"re[xyz]": lambda self: self.unit.E,
            r"t": t_unit,
            r"axis": lambda self: self.unit.length,
            r"rhobksp[1-9]": lambda self: self.unit.rho,
        }
    )

    def __init__(
        self, directory="./", append_directories=None, ad=None, inpfilename="plasma.inp"
    ):
        """EMSES出力・inpファイルを管理するオブジェクトを生成する.

        Parameters
        ----------
        directory : str or Path
            管理するディレクトリ, by default './'
        append_directories : list(str) or list(Path) or "auto"
            管理する継続ディレクトリのリスト, by default []
        ad : list(str) or list(Path) or "auto"
            管理する継続ディレクトリのリスト, by default []
        inpfilename : str, optional
            パラメータファイルの名前, by default 'plasma.inp'
        """
        if not isinstance(directory, Path):
            directory = Path(directory)
        self.directory = directory
        logger.info(
            f"Initializing Emout object for directory: {self.directory.resolve()}"
        )

        append_directories = append_directories or ad

        if append_directories == "auto":
            append_directories = self.__fetch_append_directories(directory)

        if append_directories is None:
            append_directories = []

        self.append_directories = []

        for append_directory in append_directories:
            if not isinstance(append_directory, Path):
                append_directory = Path(append_directory)
            self.append_directories.append(append_directory)

        # パラメータファイルの読み取りと単位変換器の初期化
        self.__load_inpfile(inpfilename)

    def __fetch_append_directories(self, directory: Path):
        logger.info(f"Fetching append directories for: {directory}")
        append_directories = []

        i = 2
        while True:
            path_next = f"{str(directory.resolve())}_{i}"
            directory_next = Path(path_next)

            if not directory_next.exists():
                logger.debug(f"Append directory does not exist: {directory_next}")
                break

            next_data = Emout(directory_next)
            if not next_data.is_valid():
                logger.warning(
                    f"{directory_next.resolve()} exists, but it is not valid directory."
                )
                break

            append_directories.append(directory_next)

            i += 1

        return append_directories

    def __load_inpfile(self, inpfilename: str):
        self._inp: Union[InpFile, None] = None
        self._unit: Union[Units, None] = None

        if inpfilename is None:
            return

        inpfilepath = self.directory / inpfilename

        if not inpfilepath.exists():
            return

        logger.info(f"Loading parameter file: {inpfilepath.resolve()}")
        self._inp = InpFile(inpfilepath)

        convkey = UnitConversionKey.load(inpfilepath)
        if convkey is not None:
            self._unit = Units(dx=convkey.dx, to_c=convkey.to_c)

    def __getattr__(self, __name: str) -> "GridDataSeries":
        logger.debug(f"Accessing attribute: {__name}")

        m = re.match(r"^r([eb][xyz])$", __name)
        if m:
            logger.debug(f"Relocated field requested: {m.group(1)}")
            self.__create_relocated_field_hdf5(m.group(1))

        m = re.match(r"(.+)([xyz])([xyz])$", __name)
        if m:
            logger.debug(f"Creating VectorData2d for: {__name}")
            dname = m.group(1)
            axis1 = m.group(2)
            axis2 = m.group(3)
            vector_data = VectorData2d(
                [getattr(self, f"{dname}{axis1}"), getattr(self, f"{dname}{axis2}")],
                name=__name,
            )

            setattr(self, f"_{__name}", vector_data)

            return vector_data

        filepath = self.__fetch_filepath(self.directory, f"{__name}00_0000.h5")
        logger.info(f"Loading grid data from file: {filepath.resolve()}")
        griddata = self.__load_griddata(filepath)

        for append_directory in self.append_directories:
            filepath = self.__fetch_filepath(append_directory, f"{__name}00_0000.h5")
            griddata_append = self.__load_griddata(filepath)

            griddata = griddata.chain(griddata_append)

        setattr(self, f"_{__name}", griddata)

        return griddata

    def __create_relocated_field_hdf5(self, name: str) -> None:
        axis = "zyx".index(name[-1])

        btype = ["periodic", "dirichlet", "neumann"][self.inp.mtd_vbnd[2 - axis]]

        filepath = self.__fetch_filepath(self.directory, f"{name}00_0000.h5")
        _create_relocated_field_hdf5(filepath.parent, name=name, axis=axis, btype=btype)

        for append_directory in self.append_directories:
            filepath = self.__fetch_filepath(append_directory, f"{name}00_0000.h5")
            _create_relocated_field_hdf5(
                filepath.parent, name=name, axis=axis, btype=btype
            )

    def __fetch_filepath(self, directory: Path, pattern: str) -> Path:
        filepathes = list(directory.glob(pattern))
        if len(filepathes) == 0:
            raise Exception(f"{pattern} is not found.")
        if len(filepathes) >= 2:
            raise Exception(
                f"There are multiple files that satisfy {pattern}.  Please specify so that just one is specified."
            )

        filepath = filepathes[0]

        return filepath

    def __load_griddata(self, h5file_path: Path) -> "GridDataSeries":
        if self.unit is None:
            tunit = None
            axisunit = None
        else:
            tunit = Emout.name2unit.get("t", lambda self: None)(self)
            axisunit = Emout.name2unit.get("axis", lambda self: None)(self)

        name = str(h5file_path.name).replace("00_0000.h5", "")

        if self.unit is None:
            valunit = None
        else:
            valunit = Emout.name2unit.get(name, lambda self: None)(self)

        data = GridDataSeries(
            h5file_path, name, tunit=tunit, axisunit=axisunit, valunit=valunit
        )

        return data

    @property
    def inp(self) -> Union[InpFile, None]:
        """パラメータの辞書(Namelist)を返す.

        Returns
        -------
        InpFile or None
            パラメータの辞書(Namelist)
        """
        return self._inp

    def is_valid(self) -> bool:
        """シミュレーションが正常に終了しているか判定する.

        Note:
            icurが最終ステップまで出力されているかで判定しており、hdf5ファイルだけが壊れている場合など判定が間違う場合がある。

        Returns
        -------
        bool
            シミュレーションが正常に終了している場合True
        """

        def read_last_line(file_name):
            with open(file_name, "rb") as f:
                f.seek(-2, 2)
                while f.read(1) != b"\n":
                    f.seek(-2, 1)
                return f.readline().decode("utf-8")

        if len(self.append_directories) > 0:
            dirpath = self.append_directories[-1]
        else:
            dirpath = self.directory

        if not (dirpath / "icur").exists():
            return False

        try:
            last_line = read_last_line(dirpath / "icur")
        except OSError:
            return False

        inp = InpFile(dirpath / "plasma.inp")

        return int(last_line.split()[0]) == int(inp.nstep)

    @property
    def unit(self) -> Union[Units, None]:
        """単位変換オブジェクトを返す.

        Returns
        -------
        Units or None
            単位変換オブジェクト
        """
        return self._unit

    @property
    def icur(self) -> pd.DataFrame:

        names = []
        for ispec in range(self.inp.nspec):
            names.append(f"{ispec+1}_step")
            for ipc in range(self.inp.npc):
                names.append(f"{ispec+1}_body{ipc+1}")
                names.append(f"{ispec+1}_body{ipc+1}_ema")

        df = pd.read_csv(self.directory / "icur", sep="\s+", header=None, names=names)

        return df

    @property
    def pbody(self) -> pd.DataFrame:
        names = ["step"] + [f"body{i+1}" for i in range(self.inp.npc + 1)]

        df = pd.read_csv(self.directory / "pbody", sep="\s+", names=names)

        return df


def _create_relocated_field_hdf5(
    directory: Path,
    name: str,
    axis: int,
    btype: Literal["periodic", "dirichlet", "neumann"],
):
    if name.startswith("b"):
        relocated = relocated_magnetic_field
    elif name.startswith("e"):
        relocated = relocated_electric_field

    input_filepath = directory / f"{name}00_0000.h5"
    output_filepath = directory / f"r{name}00_0000.h5"

    if output_filepath.exists():
        logger.info(f"File already exists: {output_filepath.resolve()}")
        return

    logger.info(
        f"Relocated field file not found. Creating a new file.: {output_filepath.resolve()}"
    )

    with h5py.File(input_filepath, "r") as h5_field:
        field = h5_field[name]

        with h5py.File(output_filepath, "w") as h5_relocated:
            rfield = h5_relocated.create_group(f"r{name}")

            for key in tqdm(field.keys(), desc=f"Relocating {name}"):
                rfield[key] = relocated(
                    np.array(field[key]),
                    axis=axis,
                    btype=btype,
                )

    logger.info(f"File creation completed: {output_filepath.resolve()}")


def relocated_electric_field(
    ef: np.ndarray, axis: int, btype: Literal["periodic", "dirichlet", "neumann"]
):
    def slc(a, b=None):
        s = slice(a, b) if b else a
        slices = tuple(s if i == axis else slice(None, None) for i in range(3))

        return slices

    # Relocated electric field buffer
    ref = np.zeros_like(ef)

    ref[slc(1, -1)] = 0.5 * (ef[slc(None, -2)] + ef[slc(1, -1)])

    if btype in "periodic":
        ref[slc(0)] = 0.5 * (ef[slc(-2)] + ef[slc(1)])
        ref[slc(-1)] = 0.5 * (ef[slc(-2)] + ef[slc(1)])
    elif btype in "neumann":
        ref[slc(0)] = 0
        ref[slc(-1)] = 0
    else:
        ref[slc(0)] = ef[slc(1)]
        ref[slc(-1)] = ef[slc(-2)]

    return ref


def relocated_magnetic_field(
    bf: np.array, axis: int, btype: Literal["periodic", "dirichlet", "neumann"]
):
    def slc(s1, s2=slice(None, None)):
        axis1 = (axis + 1) % 3
        axis2 = (axis + 2) % 3

        slices = [None, None, None]

        slices[axis] = slice(None, None)
        slices[axis1] = s1
        slices[axis2] = s2
        slices = tuple(slices)

        return slices

    # Relocated electric field buffer
    rbf = np.zeros_like(bf)

    # xy平面に1グリッド覆うように拡張する
    bfe = np.empty(
        np.array(bf.shape) + np.array([0 if i == axis else 1 for i in range(3)])
    )
    bfe[slc(slice(1, -1), slice(1, -1))] = bf[slc(slice(None, -1), slice(None, -1))]
    if btype in "periodic":
        bfe[slc(slice(1, -1), 0)] = bfe[slc(slice(1, -1), -2)]
        bfe[slc(slice(1, -1), -1)] = bfe[slc(slice(1, -1), 1)]
    elif btype in "dirichlet":
        bfe[slc(slice(1, -1), 0)] = -bfe[slc(slice(1, -1), 1)]
        bfe[slc(slice(1, -1), -1)] = -bfe[slc(slice(1, -1), -2)]
    else:  # if btype in "neumann":
        bfe[slc(slice(1, -1), 0)] = bfe[slc(slice(1, -1), 1)]
        bfe[slc(slice(1, -1), -1)] = bfe[slc(slice(1, -1), -2)]

    if btype in "periodic":
        bfe[slc(0)] = bfe[slc(-2)]
        bfe[slc(-1)] = bfe[slc(1)]
    elif btype in "dirichlet":
        bfe[slc(0)] = -bfe[slc(1)]
        bfe[slc(-1)] = -bfe[slc(-2)]
    else:  # if btype in "neumann":
        bfe[slc(0)] = bfe[slc(1)]
        bfe[slc(-1)] = bfe[slc(-2)]

    rbf[:, :, :] = 0.25 * (
        bfe[slc(slice(None, -1), slice(None, -1))]
        + bfe[slc(slice(1, None), slice(None, -1))]
        + bfe[slc(slice(None, -1), slice(1, None))]
        + bfe[slc(slice(1, None), slice(1, None))]
    )

    return rbf



import re
from pathlib import Path

import numpy as np
import scipy.interpolate as interp
from matplotlib.animation import PillowWriter, writers


def interp2d(mesh, n, **kwargs):
    ny, nx = mesh.shape

    if (mesh == mesh[0, 0]).all():
        return np.zeros((int(ny * n), int(nx * n))) + mesh[0, 0]

    x_sparse = np.linspace(0, 1, nx)
    y_sparse = np.linspace(0, 1, ny)

    X_sparse, Y_sparse = np.meshgrid(x_sparse, y_sparse)

    x_dense = np.linspace(0, 1, int(nx * n))
    y_dense = np.linspace(0, 1, int(ny * n))
    X_dense, Y_dense = np.meshgrid(x_dense, y_dense)

    points = (X_sparse.flatten(), Y_sparse.flatten())
    value = mesh.flatten()
    points_dense = (X_dense.flatten(), Y_dense.flatten())

    mesh_dense = interp.griddata(points, value, points_dense, **kwargs)

    return mesh_dense.reshape(X_dense.shape)


def slice2tuple(slice_obj: slice):
    """スライスオブジェクトをタプルに変換する.

    Parameters
    ----------
    slice_obj : slice
        スライスオブジェクト

    Returns
    -------
    (start, stop, step) : int
        スライス情報をもつタプル
    """
    start = slice_obj.start
    stop = slice_obj.stop
    step = slice_obj.step
    return (start, stop, step)


def range_with_slice(slice_obj, maxlen):
    """スライスを引数とするrange関数.

    Parameters
    ----------
    slice_obj : slice
        スライスオブジェクト
    maxlen : int
        最大数(スライスの値が負である場合に用いる)

    Returns
    -------
    generator
        rangeジェネレータ
    """
    start = slice_obj.start or 0
    if start < 0:
        start = maxlen + start

    stop = slice_obj.stop or maxlen
    if stop < 0:
        stop = maxlen + stop

    step = slice_obj.step or 1
    return range(start, stop, step)


class RegexDict(dict):
    """正規表現をキーとする辞書クラス."""

    def __getitem__(self, key):
        if super().__contains__(key):
            return super().__getitem__(key)

        for regex in self:
            if re.fullmatch(regex, key):
                return self[regex]

        raise KeyError()

    def __contains__(self, key):
        if super().__contains__(key):
            return True

        for regex in self:
            if re.fullmatch(regex, key):
                return True

        return False

    def get(self, key, default=None):
        try:
            return self[key]
        except Exception:
            return default


class DataFileInfo:
    """データファイル情報を管理するクラス."""

    def __init__(self, filename):
        """データファイル情報を管理するオブジェクトを生成する.

        Parameters
        ----------
        filename : str or Path
            ファイル名
        """
        if not isinstance(filename, Path):
            filename = Path(filename)
        self._filename = filename

    @property
    def filename(self):
        """ファイル名を返す.

        Returns
        -------
        Path
            ファイル名
        """
        return self._filename

    @property
    def directory(self):
        """ディレクトリの絶対パスを返す.

        Returns
        -------
        Path
            ディレクトリの絶対パス
        """
        return (self._filename / "../").resolve()

    @property
    def abspath(self):
        """ファイルの絶対パスを返す.

        Returns
        -------
        Path
            ファイルの絶対パス
        """
        return self._filename.resolve()

    def __str__(self):
        return str(self._filename)


@writers.register("quantized-pillow")
class QuantizedPillowWriter(PillowWriter):
    """色数を256としたPillowWriterラッパークラス."""

    def grab_frame(self, **savefig_kwargs):
        super().grab_frame(**savefig_kwargs)
        self._frames[-1] = self._frames[-1].convert("RGB").quantize()


def hole_mask(inp, reverse=False):
    shape = (inp.nz + 1, inp.ny + 1, inp.nx + 1)
    xl = int(inp.xlrechole[0])
    xu = int(inp.xurechole[0])
    yl = int(inp.ylrechole[0])
    yu = int(inp.yurechole[0])
    zu = int(inp.zssurf)
    zl = int(inp.zlrechole[1])

    mask = np.ones(shape, dtype=bool)
    mask[zu:, :, :] = False
    mask[zl:zu, yl : yu + 1, xl : xu + 1] = False
    return (not reverse) == mask



from abc import ABCMeta, abstractmethod
from functools import partial
from typing import Callable, List, Tuple

import numpy as np
import scipy.constants as cn
import scipy.fft


def poisson(
    rho: np.ndarray,
    dx: float,
    boundary_types: List[str] = ["periodic", "periodic", "periodic"],
    boundary_values: Tuple[Tuple[float]] = [(0.0, 0.0), (0.0, 0.0), (0.0, 0.0)],
    btypes: str = None,
    epsilon_0=cn.epsilon_0,
):
    """Solve Poisson's equation with FFT.

    Parameters
    ----------
    rho : np.ndarray
        3-dimentional array of the charge density [C/m^3]. The shape is (nz+1, ny+1, nx+1).
    boundary_types : List[str] of {'periodic', 'dirichlet', 'neumann'},
        the boundary condition types, by default ['periodic', 'periodic', 'periodic']
    boundary_values : List[Tuple[float]]
        the boundary values [(x-lower, x-upper), (y-lower, y-upper), (z-lower, z-upper)],
        by default [(0., 0.), (0., 0.), (0., 0.)]
    btypes : str
        string consisting of prefixes of boundary conditions, by default None.
        If this is set, it takes precedence over boundary_types.
    dx : float, optional
        the grid width [m], by default 1.0
    epsilon_0 : _type_, optional
        the electric constant (vacuum permittivity) [F/m], by default cn.epsilon_0

    Returns
    -------
    np.ndarray
        3-dimentional of the potential [V].

    """
    # If a boundary condition is specified in abbreviated form by btypes, revert to the original notation.
    if btypes:
        btypes_dict = {
            "p": "periodic",
            "d": "dirichlet",
            "n": "neumann",
        }
        boundary_types = [btypes_dict[btype] for btype in btypes]

    POISSON_BOUNDARIES = {
        "periodic": PeriodicPoissonBoundary,
        "dirichlet": DirichletPoissonBoundary,
        "neumann": NeumannPoissonBoundary,
    }

    # [x-boundary, y-boundary, z-boundary]
    boundaries: List[PoissonBoundary] = [
        POISSON_BOUNDARIES[_type](2 - i, boundary_values[i])
        for i, _type in enumerate(boundary_types)
    ]

    rho_target = rho[
        tuple(boundary.get_target_slice() for boundary in reversed(boundaries))
    ].copy()

    # Poisson's equation: dphi/dx^2 = -rho/epsilon_0
    rho_target = -rho_target / epsilon_0 * dx * dx

    # Transpose boundary values.
    for boundary in boundaries:
        boundary.transpose_boundary_values(rho_target, dx)

    # Create a FFT-solver with 3d data.
    forwards = [boundary.fft_forward for boundary in boundaries]
    backwards = [boundary.fft_backward for boundary in boundaries]
    fft3d = FFT3d(forwards, backwards)

    # FFT forward.
    rhok = fft3d.forward(rho_target)

    # Caluculate a modified wave number.
    modified_wave_number = np.zeros_like(rhok, dtype=float)
    nz, ny, nx = np.array(rho.shape) - 1

    for kx in range(rhok.shape[2]):
        modified_wave_number[:, :, kx] += boundaries[0].modified_wave_number(kx, nx)
    for ky in range(rhok.shape[1]):
        modified_wave_number[:, ky, :] += boundaries[1].modified_wave_number(ky, ny)
    for kz in range(rhok.shape[0]):
        modified_wave_number[kz, :, :] += boundaries[2].modified_wave_number(kz, nz)

    # Solve the equation in the wavenumber domain
    phik = rhok / modified_wave_number

    # When all boundary conditions are periodic|neumann boundaries,
    # there is no reference for the potential and it is not uniquely determined,
    # so the average is set to zero.
    if all([_type in ("periodic", "neumann") for _type in boundary_types]):
        phik[0, 0, 0] = 0.0

    # FFT backward
    _phi = fft3d.backward(phik)

    # Create an array of the same shape as the input rho array.
    phi = np.zeros_like(rho)
    phi[tuple(boundary.get_target_slice() for boundary in reversed(boundaries))] = _phi

    # In the above, the operation was performed on the array excluding the boundary values,
    # so the boundary values are substituted here.
    for boundary in boundaries:
        boundary.correct_boundary_values(phi)

    return phi


class FFT3d:
    def __init__(
        self,
        forwards: List[Callable[[np.ndarray], np.ndarray]],
        backwards: List[Callable[[np.ndarray], np.ndarray]],
    ):
        self.__forwards = forwards
        self.__backwards = backwards

    def forward(self, data3d: np.ndarray) -> np.ndarray:
        result3d = data3d

        result3d = self.__forwards[2](result3d, axis=0, norm="ortho")
        result3d = self.__forwards[1](result3d, axis=1, norm="ortho")
        result3d = self.__forwards[0](result3d, axis=2, norm="ortho")

        return result3d

    def backward(self, data3d: np.ndarray) -> np.ndarray:
        result3d = data3d

        result3d = self.__backwards[2](result3d, axis=0, norm="ortho")
        result3d = self.__backwards[1](result3d, axis=1, norm="ortho")
        result3d = self.__backwards[0](result3d, axis=2, norm="ortho")

        return result3d


class PoissonBoundary(metaclass=ABCMeta):
    def __init__(self, axis: int, boundary_values: Tuple[float] = (0.0, 0.0)):
        self.__axis = axis
        self.__boundary_values = boundary_values

    @property
    def axis(self) -> int:
        return self.__axis

    @property
    def boundary_values(self) -> Tuple[float]:
        return self.__boundary_values

    @property
    @abstractmethod
    def fft_forward(self) -> Callable[[np.ndarray], np.ndarray]:
        pass

    @property
    @abstractmethod
    def fft_backward(self) -> Callable[[np.ndarray], np.ndarray]:
        pass

    @abstractmethod
    def get_target_slice(self) -> slice:
        pass

    @abstractmethod
    def modified_wave_number(self, k: int, n: int) -> float:
        pass

    @abstractmethod
    def transpose_boundary_values(self, rho_target: np.ndarray, dx: float) -> None:
        pass

    @abstractmethod
    def correct_boundary_values(self, phi: np.ndarray) -> None:
        pass

    def _get_slices_at(self, index_axis) -> Tuple[slice]:
        return tuple(index_axis if i == self.axis else slice(None) for i in range(3))


class PeriodicPoissonBoundary(PoissonBoundary):
    @property
    def fft_forward(self) -> Callable[[np.ndarray], np.ndarray]:
        return scipy.fft.fft

    @property
    def fft_backward(self) -> Callable[[np.ndarray], np.ndarray]:
        return scipy.fft.ifft

    def get_target_slice(self) -> slice:
        return slice(0, -1)

    def modified_wave_number(self, k: int, n: int) -> float:
        if k <= int(n / 2):
            wn = 2.0 * np.sin(np.pi * k / float(n))
        else:
            wn = 2.0 * np.sin(np.pi * (n - k) / float(n))
        wn = -wn * wn

        return wn

    def transpose_boundary_values(self, rho_target: np.ndarray, dx: float) -> None:
        pass

    def correct_boundary_values(self, phi: np.ndarray) -> None:
        phi[self._get_slices_at(-1)] = phi[self._get_slices_at(0)]


class DirichletPoissonBoundary(PoissonBoundary):
    @property
    def fft_forward(self) -> Callable[[np.ndarray], np.ndarray]:
        return partial(scipy.fft.dst, type=1)

    @property
    def fft_backward(self) -> Callable[[np.ndarray], np.ndarray]:
        return partial(scipy.fft.idst, type=1)

    def get_target_slice(self) -> slice:
        return slice(1, -1)

    def modified_wave_number(self, k: int, n: int) -> float:
        wn = 2.0 * (np.cos(np.pi * (k + 1) / float(n + 1)) - 1.0)

        return wn

    def transpose_boundary_values(self, rho_target: np.ndarray, dx: float) -> None:
        rho_target[self._get_slices_at(0)] = (
            rho_target[self._get_slices_at(0)] - self.boundary_values[0]
        )
        rho_target[self._get_slices_at(-1)] = (
            rho_target[self._get_slices_at(-1)] - self.boundary_values[1]
        )

    def correct_boundary_values(self, phi: np.ndarray) -> None:
        phi[self._get_slices_at(0)] = self.boundary_values[0]
        phi[self._get_slices_at(-1)] = self.boundary_values[1]


class NeumannPoissonBoundary(PoissonBoundary):
    @property
    def fft_forward(self) -> Callable[[np.ndarray], np.ndarray]:
        return partial(scipy.fft.dct, type=1, orthogonalize=False)

    @property
    def fft_backward(self) -> Callable[[np.ndarray], np.ndarray]:
        return partial(scipy.fft.idct, type=1, orthogonalize=False)

    def get_target_slice(self) -> slice:
        return slice(None, None)

    def modified_wave_number(self, k: int, n: int) -> float:
        wn = 2.0 * (np.cos(np.pi * (k) / float(n)) - 1.0)
        return wn

    def transpose_boundary_values(self, rho_target: np.ndarray, dx: float) -> None:
        rho_target[self._get_slices_at(0)] = (
            rho_target[self._get_slices_at(0)] - self.boundary_values[0] * dx
        )
        rho_target[self._get_slices_at(-1)] = (
            rho_target[self._get_slices_at(-1)] + self.boundary_values[1] * dx
        )

    def correct_boundary_values(self, phi: np.ndarray) -> None:
        pass



import copy

import matplotlib
import matplotlib.cm as cm
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np

import emout.utils as utils

_r = 0.9
_d = 0.5
mycmap = mcolors.LinearSegmentedColormap(
    "gray-jet",
    {
        "red": (
            (0.00, 0.2, 0.2),
            (_d * (1 - _r), 0.3, 0.3),
            (0.35 * _r + (1 - _r), 0, 0),
            (0.66 * _r + (1 - _r), 1, 1),
            (0.89 * _r + (1 - _r), 1, 1),
            (1.00, 0.5, 0.5),
        ),
        "green": (
            (0.00, 0.2, 0.2),
            (_d * (1 - _r), 0.3, 0.3),
            (0.125 * _r + (1 - _r), 0, 0),
            (0.375 * _r + (1 - _r), 1, 1),
            (0.640 * _r + (1 - _r), 1, 1),
            (0.910 * _r + (1 - _r), 0, 0),
            (1.000, 0, 0),
        ),
        "blue": (
            (0.00, 0.2, 0.2),
            (_d * (1 - _r), 0.3, 0.3),
            (0.00 * _r + (1 - _r), 0.5, 0.5),
            (0.11 * _r + (1 - _r), 1, 1),
            (0.34 * _r + (1 - _r), 1, 1),
            (0.65 * _r + (1 - _r), 0, 0),
            (1.00, 0, 0),
        ),
    },
)


def figsize_with_2d(data2d, dpi=10):
    """2次元データから図のサイズを計算する.

    Parameters
    ----------
    data2d : numpy.ndarray
        2次元データ
    dpi : int, optional
        1データを何pixelで表すか, by default 10

    Returns
    -------
    (float, float)
        図のサイズ
    """
    px = 1 / plt.rcParams["figure.dpi"] * dpi
    figsize = (data2d.shape[1] * px, data2d.shape[0] * px)
    return figsize


def plot_2dmap(
    data2d,
    mesh=None,
    savefilename=None,
    cmap=mycmap,
    mask_color="gray",
    vmin=None,
    vmax=None,
    figsize=None,
    xlabel=None,
    ylabel=None,
    title=None,
    interpolation="bilinear",
    dpi=10,
    colorbar_label="",
    cbargs={},
    add_colorbar=True,
    **kwargs
):
    """2次元カラーマップをプロットする.

    Parameters
    ----------
    data2d : numpy.ndarray
        2次元データ
    mesh : (numpy.ndarray, numpy.ndarray), optional
        メッシュ, by default None
    savefilename : str, optional
        保存するファイル名(Noneの場合保存しない), by default None
    cmap : matplotlib.Colormap or str or None, optional
        カラーマップ, by default cm.coolwarm
    mask_color : str
        マスクされた位置の色, by default 'gray'
    vmin : float, optional
        最小値, by default None
    vmax : float, optional
        最大値, by default None
    figsize : (float, float), optional
        図のサイズ, by default None
    xlabel : str, optional
        x軸のラベル, by default None
    ylabel : str, optional
        y軸のラベル, by default None
    title : str, optional
        タイトル, by default None
    interpolation : str, optional
        用いる補間方法, by default 'bilinear'
    dpi : int, optional
        解像度(figsizeが指定された場合は無視される), by default 10
    add_colorbar: bool, optional
        カラーバーを追加する場合True, by default True

    Returns
    -------
    AxesImage or None
        プロットしたimageデータ(保存した場合None)
    """
    if mesh is None:
        x = list(range(data2d.shape[1]))
        y = list(range(data2d.shape[0]))
        mesh = np.meshgrid(x, y)

    if cmap is not None:
        if isinstance(cmap, str):
            cmap = copy.copy(plt.get_cmap(str(cmap)))
        else:
            cmap = copy.copy(cmap)
        cmap.set_bad(color=mask_color)

    extent = [mesh[0][0, 0], mesh[0][-1, -1], mesh[1][0, 0], mesh[1][-1, -1]]
    img = plt.imshow(
        data2d,
        interpolation=interpolation,
        cmap=cmap,
        origin="lower",
        vmin=vmin,
        vmax=vmax,
        extent=extent,
        aspect="auto",
        **kwargs
    )
    if add_colorbar:
        if "cb" in cbargs:
            cb = plt.colorbar(label=colorbar_label, **cbargs["cb"])
        else:
            cb = plt.colorbar(label=colorbar_label)

    if "others" in cbargs and "yticklabels" in cbargs["others"]:
        cb.ax.set_yticklabels(cbargs["others"]["yticklabels"])

    if title is not None:
        plt.title(title)
    if xlabel is not None:
        plt.xlabel(xlabel)
    if ylabel is not None:
        plt.ylabel(ylabel)

    if savefilename is not None:
        plt.gcf().savefig(savefilename)
        plt.close(plt.gcf())
        return None
    else:
        return img


def plot_2d_contour(
    data2d,
    mesh=None,
    levels=None,
    colors=["black"],
    cmap=None,
    alpha=1,
    vmin=None,
    vmax=None,
    savefilename=None,
    figsize=None,
    xlabel=None,
    ylabel=None,
    title=None,
    dpi=10,
    fmt="%1.1f",
    fontsize=12,
    **kwargs
):
    """2次元等高線をプロットする.

    Parameters
    ----------
    data2d : numpy.ndarray
        2次元データ
    mesh : (numpy.ndarray, numpy.ndarray), optional
        メッシュ, by default None
    levels : int
        等高線数, by default None
    alpha : float
        透明度(0.0~1.0), by default 1
    savefilename : str, optional
        保存するファイル名(Noneの場合保存しない), by default None
    cmap : matplotlib.Colormap or str or None, optional
        カラーマップ, by default None
    mask_color : str
        マスクされた位置の色, by default 'gray'
    vmin : float, optional
        最小値, by default None
    vmax : float, optional
        最大値, by default None
    figsize : (float, float), optional
        図のサイズ, by default None
    xlabel : str, optional
        x軸のラベル, by default None
    ylabel : str, optional
        y軸のラベル, by default None
    title : str, optional
        タイトル, by default None
    interpolation : str, optional
        用いる補間方法, by default 'bilinear'
    dpi : int, optional
        解像度(figsizeが指定された場合は無視される), by default 10
    fmt : str
        clabelの形式, by default '%1.1f'
    fontsize : str
        clabelのフォントサイズ, by default 12

    Returns
    -------
    AxesImage or None
        プロットしたimageデータ(保存した場合None)
    """
    if mesh is None:
        x = list(range(data2d.shape[1]))
        y = list(range(data2d.shape[0]))
        mesh = np.meshgrid(x, y)

    kwargs = {
        "alpha": alpha,
        "vmin": vmin,
        "vmax": vmax,
    }
    if cmap is None:
        kwargs["colors"] = colors
    else:
        kwargs["cmap"] = cmap
    if levels is not None:
        kwargs["levels"] = levels
    cont = plt.contour(*mesh, data2d, **kwargs)
    cont.clabel(fmt=fmt, fontsize=fontsize)

    if title is not None:
        plt.title(title)
    if xlabel is not None:
        plt.xlabel(xlabel)
    if ylabel is not None:
        plt.ylabel(ylabel)

    if savefilename is not None:
        plt.gcf().savefig(savefilename)
        plt.close(plt.gcf())
        return None
    else:
        return cont


def plot_surface(
    x,
    y,
    z,
    value,
    ax3d=None,
    add_colorbar=False,
    savefilename=None,
    cmap=cm.jet,
    mask_color="gray",
    vmin=None,
    vmax=None,
    figsize=None,
    xlabel=None,
    ylabel=None,
    zlabel=None,
    title=None,
    ninterp=1,
    function="linear",
    dpi=10,
    colorbar_label="",
):
    """3次元表面プロットをする.

    Parameters
    ----------
    x : (numpy.ndarray, numpy.ndarray), optional
        x座標のメッシュ
    y : (numpy.ndarray, numpy.ndarray), optional
        y座標のメッシュ
    z : (numpy.ndarray, numpy.ndarray), optional
        z座標のメッシュ
    val : (numpy.ndarray, numpy.ndarray), optional
        値のメッシュ
    ax3d : Axes3D
        Axes3Dオブジェクト, by default None
    savefilename : str, optional
        保存するファイル名(Noneの場合保存しない), by default None
    cmap : matplotlib.Colormap or str or None, optional
        カラーマップ, by default cm.coolwarm
    vmin : float, optional
        最小値, by default None
    vmax : float, optional
        最大値, by default None
    figsize : (float, float), optional
        図のサイズ, by default None
    xlabel : str, optional
        x軸のラベル, by default None
    ylabel : str, optional
        y軸のラベル, by default None
    zlabel : str, optional
        z軸のラベル, by default None
    title : str, optional
        タイトル, by default None
    dpi : int, optional
        解像度(figsizeが指定された場合は無視される), by default 10

    Returns
    -------
    AxesImage or None
        プロットしたimageデータ(保存した場合None)
    """
    if savefilename is not None:
        if figsize is None:
            fig = plt.figure()
        else:
            if figsize == "auto":
                figsize = figsize_with_2d(x, dpi=dpi)
            fig = plt.figure(figsize=figsize)
    else:
        fig = plt.gcf()

    if ax3d is None:
        ax3d = fig.gca(projection="3d")

    if cmap is not None:
        if isinstance(cmap, str):
            cmap = copy.copy(cm.get_cmap(str(cmap)))
        else:
            cmap = copy.copy(cmap)
        cmap.set_bad(color=mask_color)

    if ninterp is not None:
        x = utils.interp2d(x, ninterp, method=function)
        y = utils.interp2d(y, ninterp, method=function)
        z = utils.interp2d(z, ninterp, method=function)
        value = utils.interp2d(value, ninterp)

    if vmin is None:
        vmin = value.min()
    if vmax is None:
        vmax = value.max()

    norm = matplotlib.colors.Normalize(vmin, vmax)
    mappable = cm.ScalarMappable(cmap=cmap, norm=norm)
    mappable.set_array([])
    value_colors = mappable.to_rgba(value)

    surf = ax3d.plot_surface(
        x, y, z, facecolors=value_colors, vmin=vmin, vmax=vmax, shade=False
    )
    if add_colorbar:
        plt.colorbar(mappable, ax=ax3d, label=colorbar_label)

    if title is not None:
        ax3d.set_title(title)
    if xlabel is not None:
        ax3d.set_xlabel(xlabel)
    if ylabel is not None:
        ax3d.set_ylabel(ylabel)
    if zlabel is not None:
        ax3d.set_zlabel(zlabel)

    if savefilename is not None:
        fig.savefig(savefilename)
        plt.close(fig)
        return None
    else:
        return surf


def plot_line(
    data1d,
    x=None,
    savefilename=None,
    vmin=None,
    vmax=None,
    figsize=None,
    xlabel=None,
    ylabel=None,
    label=None,
    title=None,
    **kwargs
):
    """1次元データをプロットする.

    Parameters
    ----------
    data1d : array-like or scalar
        プロットする1次元データ
    x : array-like or scalar
        横軸となる1次元データ, by default None
    savefilename : str, optional
        保存するファイル名, by default None
    vmin : float, optional
        最小値, by default None
    vmax : float, optional
        最大値, by default None
    figsize : (float, float), optional
        図のサイズ, by default None
    xlabel : str, optional
        横軸のラベル, by default None
    ylabel : str, optional
        縦軸のラベル, by default None
    label : str, optional
        ラベル, by default None
    title : str, optional
        タイトル, by default None

    Returns
    -------
    Line2D or None
        プロットデータを表す線オブジェクト(保存した場合None)
    """
    if savefilename is not None:
        if figsize is None:
            fig = plt.figure()
        else:
            fig = plt.figure(figsize=figsize)

    if x is None:
        line = plt.plot(data1d, label=label, **kwargs)
    else:
        line = plt.plot(x, data1d, label=label, **kwargs)
    plt.ylim([vmin, vmax])

    if title is not None:
        plt.title(title)
    if xlabel is not None:
        plt.xlabel(xlabel)
    if ylabel is not None:
        plt.ylabel(ylabel)

    if savefilename is not None:
        fig.savefig(savefilename)
        plt.close(fig)
        return None
    else:
        return line


def plot_2d_vector(
    x_data2d,
    y_data2d,
    mesh=None,
    savefilename=None,
    color=None,
    scale=1,
    scaler="standard",
    skip=1,
    easy_to_read=True,
    figsize=None,
    xlabel=None,
    ylabel=None,
    title=None,
    dpi=10,
    cmap=None,
):
    """2次元ベクトル図をプロットする.

    Parameters
    ----------
    x_data2d, y_data2d : numpy.ndarray
        2次元データ
    mesh : (numpy.ndarray, numpy.ndarray), optional
        メッシュ, by default None
    savefilename : str, optional
        保存するファイル名(Noneの場合保存しない), by default None
    color : str
        ベクトルの色, by default None
    scale : float
        ベクトルの大きさ係数(最終的な大きさにこの値を掛ける), by default 1
    skip : int
        プロットするデータ間隔, by default 1
    easy_to_read : bool
        ベクトルを見やすい大きさにスケーリングするならTrue, by default True
    figsize : (float, float), optional
        図のサイズ, by default None
    xlabel : str, optional
        x軸のラベル, by default None
    ylabel : str, optional
        y軸のラベル, by default None
    title : str, optional
        タイトル, by default None
    interpolation : str, optional
        用いる補間方法, by default 'bilinear'
    dpi : int, optional
        解像度(figsizeが指定された場合は無視される), by default 10

    Returns
    -------
    AxesImage or None
        プロットしたimageデータ(保存した場合None)
    """
    fig = None
    if savefilename is not None:
        if figsize is None:
            fig = plt.figure()
        else:
            if figsize == "auto":
                figsize = figsize_with_2d(x_data2d, dpi=dpi)
            fig = plt.figure(figsize=figsize)

    if mesh is None:
        x = list(range(x_data2d.shape[1]))
        y = list(range(x_data2d.shape[0]))
        mesh = np.meshgrid(x, y)

    x = mesh[0]
    y = mesh[1]
    U = np.array(x_data2d)
    V = np.array(y_data2d)

    x_skip = skip if type(skip) == int else skip[0]
    y_skip = skip if type(skip) == int else skip[1]
    x = x[::y_skip, ::x_skip]
    y = y[::y_skip, ::x_skip]
    U = U[::y_skip, ::x_skip]
    V = V[::y_skip, ::x_skip]

    norm = np.sqrt(U**2 + V**2)

    if scaler == "standard":
        norm_max = np.nanmax(np.abs(norm))
        U /= norm_max
        V /= norm_max

    elif scaler == "normal":
        U /= norm
        V /= norm

    elif scaler == "log":
        U = U / norm * np.log(norm + 1)
        V = V / norm * np.log(norm + 1)

    # 見やすい大きさに線形スケーリングを行う
    if easy_to_read:
        dx = (x.max() - x.min()) / x.shape[0]
        multiplier = dx * 1.2
        norm_mean = np.nanmean(np.sqrt(U**2 + V**2))
        U *= scale / norm_mean * multiplier
        V *= scale / norm_mean * multiplier

    if cmap is None:
        img = plt.quiver(
            x,
            y,
            U,
            V,
            angles="xy",
            scale_units="xy",
            scale=1,
        )
    else:
        img = plt.quiver(
            x,
            y,
            U,
            V,
            np.sqrt(U**2 + V**2),
            angles="xy",
            scale_units="xy",
            scale=1,
            cmap=cmap,
        )

    if title is not None:
        plt.title(title)
    if xlabel is not None:
        plt.xlabel(xlabel)
    if ylabel is not None:
        plt.ylabel(ylabel)

    if savefilename is not None:
        fig.savefig(savefilename)
        plt.close(fig)
        return None
    else:
        return img


def plot_2d_streamline(
    x_data2d,
    y_data2d,
    mesh=None,
    savefilename=None,
    skip=1,
    figsize=None,
    xlabel=None,
    ylabel=None,
    title=None,
    dpi=10,
    color=None,
    cmap=None,
    norm="linear",
    vmin=None,
    vmax=None,
    density=1,
):
    """2次元ベクトル図をプロットする.

    Parameters
    ----------
    x_data2d, y_data2d : numpy.ndarray
        2次元データ
    mesh : (numpy.ndarray, numpy.ndarray), optional
        メッシュ, by default None
    savefilename : str, optional
        保存するファイル名(Noneの場合保存しない), by default None
    color : str
        ベクトルの色, by default None
    scale : float
        ベクトルの大きさ係数(最終的な大きさにこの値を掛ける), by default 1
    skip : int
        プロットするデータ間隔, by default 1
    easy_to_read : bool
        ベクトルを見やすい大きさにスケーリングするならTrue, by default True
    figsize : (float, float), optional
        図のサイズ, by default None
    xlabel : str, optional
        x軸のラベル, by default None
    ylabel : str, optional
        y軸のラベル, by default None
    title : str, optional
        タイトル, by default None
    interpolation : str, optional
        用いる補間方法, by default 'bilinear'
    dpi : int, optional
        解像度(figsizeが指定された場合は無視される), by default 10

    Returns
    -------
    AxesImage or None
        プロットしたimageデータ(保存した場合None)
    """
    fig = None
    if savefilename is not None:
        if figsize is None:
            fig = plt.figure()
        else:
            if figsize == "auto":
                figsize = figsize_with_2d(x_data2d, dpi=dpi)
            fig = plt.figure(figsize=figsize)

    if mesh is None:
        x = list(range(x_data2d.shape[1]))
        y = list(range(x_data2d.shape[0]))
        mesh = np.meshgrid(x, y)

    x = mesh[0]
    y = mesh[1]
    U = np.array(x_data2d)
    V = np.array(y_data2d)

    x_skip = skip if type(skip) == int else skip[0]
    y_skip = skip if type(skip) == int else skip[1]
    x = x[::y_skip, ::x_skip]
    y = y[::y_skip, ::x_skip]
    U = U[::y_skip, ::x_skip]
    V = V[::y_skip, ::x_skip]

    if cmap:
        length = np.sqrt(U**2 + V**2)
        vmin = vmin or length.min()
        vmax = vmax or length.max()

        if norm == "linear":
            norm = mcolors.Normalize(vmin=vmin, vmax=vmax)
        elif norm == "log":
            norm = mcolors.LogNorm(vmin=vmin, vmax=vmax)
        elif norm == "centered":
            norm = mcolors.CenteredNorm()
        elif norm == "symlog":
            norm = mcolors.SymLogNorm(vmin=vmin, vmax=vmax)

        img = plt.streamplot(
            x,
            y,
            U,
            V,
            color=length,
            cmap=cmap,
            norm=norm,
            density=density,
        )
    else:
        img = plt.streamplot(
            x,
            y,
            U,
            V,
            color=color,
            density=density,
        )

    if title is not None:
        plt.title(title)
    if xlabel is not None:
        plt.xlabel(xlabel)
    if ylabel is not None:
        plt.ylabel(ylabel)

    if savefilename is not None:
        fig.savefig(savefilename)
        plt.close(fig)
        return None
    else:
        return img


def plot_3d_quiver(
    x_data3d,
    y_data3d,
    z_data3d,
    ax3d=None,
    mesh=None,
    savefilename=None,
    color=None,
    scale=1,
    scaler="standard",
    skip=1,
    easy_to_read=True,
    figsize=None,
    xlabel=None,
    ylabel=None,
    title=None,
    dpi=10,
    cmap=None,
):
    """2次元ベクトル図をプロットする.

    Parameters
    ----------
    x_data2d, y_data2d : numpy.ndarray
        2次元データ
    mesh : (numpy.ndarray, numpy.ndarray), optional
        メッシュ, by default None
    savefilename : str, optional
        保存するファイル名(Noneの場合保存しない), by default None
    color : str
        ベクトルの色, by default None
    scale : float
        ベクトルの大きさ係数(最終的な大きさにこの値を掛ける), by default 1
    skip : int
        プロットするデータ間隔, by default 1
    easy_to_read : bool
        ベクトルを見やすい大きさにスケーリングするならTrue, by default True
    figsize : (float, float), optional
        図のサイズ, by default None
    xlabel : str, optional
        x軸のラベル, by default None
    ylabel : str, optional
        y軸のラベル, by default None
    title : str, optional
        タイトル, by default None
    interpolation : str, optional
        用いる補間方法, by default 'bilinear'
    dpi : int, optional
        解像度(figsizeが指定された場合は無視される), by default 10

    Returns
    -------
    AxesImage or None
        プロットしたimageデータ(保存した場合None)
    """
    fig = None
    if savefilename is not None:
        if figsize is None:
            fig = plt.figure()
        else:
            if figsize == "auto":
                figsize = figsize_with_2d(x_data3d[:, 0, :], dpi=dpi)
            fig = plt.figure(figsize=figsize)

    if ax3d is None:
        ax3d = fig.add_subplot(projection="3d")

    if mesh is None:
        x = list(range(x_data3d.shape[1]))
        y = list(range(x_data3d.shape[0]))
        z = list(range(x_data3d.shape[0]))
        mesh = np.meshgrid(x, y, z)

    x = mesh[0]
    y = mesh[1]
    z = mesh[2]
    U = np.array(x_data3d)
    V = np.array(y_data3d)
    W = np.array(z_data3d)

    x_skip = skip if type(skip) == int else skip[0]
    y_skip = skip if type(skip) == int else skip[1]
    z_skip = skip if type(skip) == int else skip[2]
    x = x[::z_skip, ::y_skip, ::x_skip]
    y = y[::z_skip, ::y_skip, ::x_skip]
    z = z[::z_skip, ::y_skip, ::x_skip]
    U = U[::z_skip, ::y_skip, ::x_skip]
    V = V[::z_skip, ::y_skip, ::x_skip]
    W = W[::z_skip, ::y_skip, ::x_skip]

    norm = np.sqrt(U**2 + V**2)

    if scaler == "standard":
        norm_max = np.nanmax(np.abs(norm))
        U /= norm_max
        V /= norm_max
        W /= norm_max

    elif scaler == "normal":
        U /= norm
        V /= norm
        W /= norm

    elif scaler == "log":
        U = U / norm * np.log(norm + 1)
        V = V / norm * np.log(norm + 1)
        W = W / norm * np.log(norm + 1)

    # 見やすい大きさに線形スケーリングを行う
    if easy_to_read:
        dx = (x.max() - x.min()) / x.shape[0]
        multiplier = dx * 1.2
        norm_mean = np.nanmean(np.sqrt(U**2 + V**2))
        U *= scale / norm_mean * multiplier
        V *= scale / norm_mean * multiplier
        W *= scale / norm_mean * multiplier

    if cmap is None:
        img = ax3d.quiver(
            x,
            y,
            z,
            U,
            V,
            W,
            angles="xy",
            scale_units="xy",
            scale=1,
        )
    else:
        img = ax3d.quiver(
            x,
            y,
            z,
            U,
            V,
            W,
            np.sqrt(U**2 + V**2 + W**2),
            angles="xy",
            scale_units="xy",
            scale=1,
            cmap=cmap,
        )

    if title is not None:
        plt.title(title)
    if xlabel is not None:
        plt.xlabel(xlabel)
    if ylabel is not None:
        plt.ylabel(ylabel)

    if savefilename is not None:
        fig.savefig(savefilename)
        plt.close(fig)
        return None
    else:
        return img



import collections
import warnings
from os import PathLike
from typing import Callable, List, Tuple, Union, Literal

import matplotlib.animation as animation
import matplotlib.pyplot as plt
import numpy as np

import emout.utils as utils


def flatten_list(l):
    for el in l:
        if isinstance(el, collections.abc.Iterable) and not isinstance(
            el, (str, bytes)
        ):
            yield from flatten_list(el)
        else:
            yield el


ANIMATER_PLOT_MODE = Literal["return", "show", "to_html", "save"]


class Animator:
    def __init__(
        self,
        layout: List[List[List[Union["FrameUpdater", Callable[[int], None], None]]]],
    ):
        self._layout = layout

    def plot(
        self,
        fig: Union[plt.Figure, None] = None,
        action: ANIMATER_PLOT_MODE = "to_html",
        filename: PathLike = None,
        interval: int = 200,
        repeat: bool = True,
        show: bool = False,
        savefilename: PathLike = None,
        to_html: bool = False,
    ):
        """gifアニメーションを作成する

        Parameters
        ----------
        fig : Figure
            アニメーションを描画するFigure(Noneの場合新しく作成する), by default None

        action : {'return', 'show', 'to_html', 'save'}, optional, by default 'to_html'
            Determines the behavior of the function:

            - 'return': The plot object (fig, ani) is returned without rendering it.
            - 'show': The plot is displayed immediately.
            - 'to_html': The plot is converted to an Ipython.display.HTML object and returned.
            - 'save': The plot is saved to a file specified by 'filename' argument.

        filename : str, optional
            保存するファイル名(actionが'save'以外の場合やNoneの場合保存されない), by default None

        interval : int, optional
            フレーム間のインターバル(ミリ秒), by default 400
        repeat : bool
            アニメーションをループするならTrue, by default True

        show : bool, optional
            プロットを表示する場合True(ファイルに保存する場合は非表示), by default

            .. deprecated :: 1.2.1

               This parameter is deprecated and will be removed in version 2.0.0.
               Use the 'action'='show' instead for equivalent functionality.

        savefilename : str, optional
            保存するファイル名(Noneの場合保存しない), by default None

            .. deprecated :: 1.2.1

               This parameter is deprecated and will be removed in version 2.0.0.
               Use the plot('action'='save', filename='example.gif') instead for equivalent functionality.

        to_html : bool
            アニメーションをHTMLとして返す. (使用例: Jupyter Notebook等でアニメーションを描画する際等)

            .. deprecated :: 1.2.1

               This parameter is deprecated and will be removed in version 2.0.0.
               Use the 'action'='to_html' instead for equivalent functionality.

        Returns
        -------
        Depending on the selected action:

        - If 'return': Returns the tuple of the plot object (fig, animation).
        - If 'show': Does not return anything, displays the plot.
        - If 'to_html': Returns an Ipython.display.HTML object of the plot (for Jupyter).
        - If 'save': Does not return anything, saves the plot to a file.

        Examples
        --------
        >>> fig, ani = plot(action="return")
        Returns the tuple of the plot object.

        >>> plot(action="show")
        Displays the plot.

        >>> html = plot(action="to_html")
        Returns the HTML representation of the plot.

        >>> plot(action="save", filename = "example.gif")
        Saves the plot to a file.
        """
        if show:
            warnings.warn(
                "The 'show' flag is deprecated. "
                "Please use gifplot(action='show') instead.",
                DeprecationWarning,
            )
            action = "show"

        if to_html:
            warnings.warn(
                "The 'to_html' flag is deprecated. "
                "Please use gifplot(action='to_html') instead.",
                DeprecationWarning,
            )
            action = "to_html"

        if savefilename:
            warnings.warn(
                "The 'savefilename' argument is scheduled to change. "
                "Please use gifplot(action='save', filename='example.gif'), instead",
                DeprecationWarning,
            )
            action = "save"
            filename = savefilename

        if fig is None:
            fig = plt.gcf()

        def _update_all(i):
            plt.clf()
            j = 0
            shape = self.shape
            for line in self._layout:
                for plot in line:
                    j += 1

                    if plot[0] is None:
                        continue

                    plt.subplot(shape[0], shape[1], j)
                    for updater in plot:
                        if updater is None:
                            continue
                        updater(i)

        frames = self.frames

        ani = animation.FuncAnimation(
            fig,
            _update_all,
            interval=interval,
            frames=frames,
            repeat=repeat,
        )

        if action == "to_html":
            from IPython.display import HTML

            return HTML(ani.to_jshtml())
        elif action == "save" and (filename is not None):
            ani.save(filename, writer="quantized-pillow")
        elif action == "show":
            plt.show()
        else:
            return fig, ani

    @property
    def frames(self):
        """管理いているFrameUpdaterの最小フレーム数."""
        updaters = list(flatten_list(self._layout))
        if not updaters:
            raise ValueError("Updaters have no elements")

        # フレーム数の最小値を返す
        frames = min(
            len(updater) for updater in updaters if isinstance(updater, FrameUpdater)
        )
        return frames

    @property
    def shape(self):
        """レイアウトの形状."""
        nrows = len(self._layout)

        ncols = 1
        for l in self._layout:
            ncols = max(ncols, len(l))

        return (nrows, ncols)


class FrameUpdater:
    def __init__(
        self,
        data,
        axis: int = 0,
        title: Union[str, None] = None,
        notitle: bool = False,
        offsets: Union[
            Tuple[Union[float, str], Union[float, str], Union[float, str]], None
        ] = None,
        use_si: bool = True,
        vmin: float = None,
        vmax: float = None,
        **kwargs,
    ):
        if data.valunit is None:
            use_si = False

        if title is None:
            title = data.name

        if use_si:
            vmin = vmin or data.valunit.reverse(data.min())
            vmax = vmax or data.valunit.reverse(data.max())
        else:
            vmin = vmin or data.min()
            vmax = vmax or data.max()

        self.data = data
        self.axis = axis
        self.title = title
        self.notitle = notitle
        self.offsets = offsets
        self.use_si = use_si
        self.vmin = vmin
        self.vmax = vmax
        self.kwargs = kwargs

    def __call__(self, i: int):
        self.update(i)

    def update(self, i: int):
        data = self.data
        axis = self.axis
        title = self.title
        notitle = self.notitle
        offsets = self.offsets
        use_si = self.use_si
        vmin = self.vmin
        vmax = self.vmax
        kwargs = self.kwargs

        # 指定した軸でスライス
        slices = [slice(None)] * len(data.shape)
        slices[axis] = i
        val = data[tuple(slices)]

        # タイトルの設定
        if notitle:
            _title = title if len(title) > 0 else None
        else:
            ax = data.slice_axes[axis]
            slc = data.slices[ax]
            maxlen = data.shape[axis]

            line = np.array(utils.range_with_slice(slc, maxlen=maxlen), dtype=float)

            if offsets is not None:
                line = self._offseted(line, offsets[0])

            index = line[i]

            if use_si:  # SI単位系を用いる場合
                axisunit = data.axisunits[ax]
                _title = f"{title}({axisunit.reverse(index):.4e} {axisunit.unit}"

            else:  # EMSES単位系を用いる場合
                _title = f"{title}({index})"

        if offsets is not None:
            offsets2d = offsets[1:]
        else:
            offsets2d = None

        val.plot(
            vmin=vmin,
            vmax=vmax,
            title=_title,
            use_si=use_si,
            offsets=offsets2d,
            **kwargs,
        )

    def _offseted(self, line: List, offset: Union[str, float]):
        if offset == "left":
            line -= line[0]
        elif offset == "center":
            line -= line[len(line) // 2]
        elif offset == "right":
            line -= line[-1]
        else:
            line += offset
        return line

    def to_animator(self, layout=None):
        """アニメーターに変換する.

        Parameters
        ----------
        layout: List[List[List[FrameUpdater]]]
            アニメーションプロットのレイアウト
        """
        if layout is None:
            layout = [[[self]]]

        return Animator(layout=layout)

    def __len__(self):
        return self.data.shape[self.axis]
